import os
import tensorflow as tf


def setup_threads(num_threads=os.cpu_count() - 1):
    num_threads_str = str(num_threads)
    os.environ["OMP_NUM_THREADS"] = num_threads_str
    os.environ["TF_NUM_INTRAOP_THREADS"] = num_threads_str
    os.environ["TF_NUM_INTEROP_THREADS"] = num_threads_str
    tf.config.threading.set_inter_op_parallelism_threads(
        num_threads
    )
    tf.config.threading.set_intra_op_parallelism_threads(
        num_threads
    )
    tf.config.set_soft_device_placement(True)


import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.base import BaseEstimator


__all__ = [
    'HistogramCalibrator',
    'PlattCalibrator',
    'PlattHistogramCalibrator'
]

from exoml.ml.calibration.utils import create_binned_data, get_bin_boundaries


#File taken from https://github.com/ethen8181/machine-learning/blob/master/model_selection/prob_calibration/calibration_module/utils.py
class HistogramCalibrator(BaseEstimator):
    """
    Bins the data based on equal size interval (each bin contains approximately
    equal size of samples).

    Parameters
    ----------
    n_bins : int, default 15
        A bigger bin number requires more data. In general,
        the larger the bin size, the closer the calibration error
        will be to the true calibration error.

    Attributes
    ----------
    bins_ : 1d ndarray
        Boundaries for each bin.

    bins_score_ : 1d ndarray
        Calibration score for each bin.
    """

    def __init__(self, n_bins: int=15):
        self.n_bins = n_bins

    def fit(self, y_prob: np.ndarray, y_true: np.ndarray):
        """
        Learns the bin boundaries and calibration score for each bin.

        Parameters
        ----------
        y_prob : 1d ndarray
            Raw probability/score of the positive class.

        y_true : 1d ndarray
            Binary true targets.

        Returns
        -------
        self
        """
        binned_y_true, binned_y_prob = create_binned_data(y_true, y_prob, self.n_bins)
        self.bins_ = get_bin_boundaries(binned_y_prob)
        self.bins_score_ = np.array([np.mean(value) for value in binned_y_true])
        return self

    def predict(self, y_prob: np.ndarray) -> np.ndarray:
        """
        Predicts the calibrated probability.

        Parameters
        ----------
        y_prob : 1d ndarray
            Raw probability/score of the positive class.

        Returns
        -------
        y_calibrated_prob : 1d ndarray
            Calibrated probability.
        """
        indices = np.searchsorted(self.bins_, y_prob)
        return self.bins_score_[indices]


class PlattCalibrator(BaseEstimator):
    """
    Boils down to applying a Logistic Regression.

    Parameters
    ----------
    log_odds : bool, default True
        Logistic Regression assumes a linear relationship between its input
        and the log-odds of the class probabilities. Converting the probability
        to log-odds scale typically improves performance.

    Attributes
    ----------
    coef_ : ndarray of shape (1,)
        Binary logistic regression's coefficient.

    intercept_ : ndarray of shape (1,)
        Binary logistic regression's intercept.
    """

    def __init__(self, log_odds: bool=True):
        self.log_odds = log_odds

    def fit(self, y_prob: np.ndarray, y_true: np.ndarray):
        """
        Learns the logistic regression weights.

        Parameters
        ----------
        y_prob : 1d ndarray
            Raw probability/score of the positive class.

        y_true : 1d ndarray
            Binary true targets.

        Returns
        -------
        self
        """
        self.fit_predict(y_prob, y_true)
        return self

    @staticmethod
    def _convert_to_log_odds(y_prob: np.ndarray) -> np.ndarray:
        eps = 1e-12
        y_prob = np.clip(y_prob, eps, 1 - eps)
        y_prob = np.log(y_prob / (1 - y_prob))
        return y_prob

    def predict(self, y_prob: np.ndarray) -> np.ndarray:
        """
        Predicts the calibrated probability.

        Parameters
        ----------
        y_prob : 1d ndarray
            Raw probability/score of the positive class.

        Returns
        -------
        y_calibrated_prob : 1d ndarray
            Calibrated probability.
        """
        if self.log_odds:
            y_prob = self._convert_to_log_odds(y_prob)

        output = self._transform(y_prob)
        return output

    def _transform(self, y_prob: np.ndarray) -> np.ndarray:
        output = y_prob * self.coef_[0] + self.intercept_
        output = 1 / (1 + np.exp(-output))
        return output

    def fit_predict(self, y_prob: np.ndarray, y_true: np.ndarray) -> np.ndarray:
        """
        Chain the .fit and .predict step together.

        Parameters
        ----------
        y_prob : 1d ndarray
            Raw probability/score of the positive class.

        y_true : 1d ndarray
            Binary true targets.

        Returns
        -------
        y_calibrated_prob : 1d ndarray
            Calibrated probability.
        """
        if self.log_odds:
            y_prob = self._convert_to_log_odds(y_prob)

        # the class expects 2d ndarray as input features
        logistic = LogisticRegression(C=1e10, solver='lbfgs')
        logistic.fit(y_prob.reshape(-1, 1), y_true)
        self.coef_ = logistic.coef_[0]
        self.intercept_ = logistic.intercept_

        y_calibrated_prob = self._transform(y_prob)
        return y_calibrated_prob


class PlattHistogramCalibrator(PlattCalibrator):
    """
    Boils down to first applying a Logistic Regression then perform
    histogram binning.

    Parameters
    ----------
    log_odds : bool, default True
        Logistic Regression assumes a linear relationship between its input
        and the log-odds of the class probabilities. Converting the probability
        to log-odds scale typically improves performance.

    n_bins : int, default 15
        A bigger bin number requires more data. In general,
        the larger the bin size, the closer the calibration error
        will be to the true calibration error.

    Attributes
    ----------
    coef_ : ndarray of shape (1,)
        Binary logistic regresion's coefficient.

    intercept_ : ndarray of shape (1,)
        Binary logistic regression's intercept.

    bins_ : 1d ndarray
        Boundaries for each bin.

    bins_score_ : 1d ndarray
        Calibration score for each bin.
    """

    def __init__(self, log_odds: bool=True, n_bins: int=15):
        super().__init__(log_odds)
        self.n_bins = n_bins

    def fit(self, y_prob: np.ndarray, y_true: np.ndarray):
        """
        Learns the logistic regression weights and the
        bin boundaries and calibration score for each bin.

        Parameters
        ----------
        y_prob : 1d ndarray
            Raw probability/score of the positive class.

        y_true : 1d ndarray
            Binary true targets.

        Returns
        -------
        self
        """
        y_prob_platt = super().fit_predict(y_prob, y_true)
        binned_y_true, binned_y_prob = create_binned_data(y_true, y_prob_platt, self.n_bins)
        self.bins_ = get_bin_boundaries(binned_y_prob)
        self.bins_score_ = np.array([np.mean(value) for value in binned_y_prob])
        return self

    def predict(self, y_prob: np.ndarray) -> np.ndarray:
        """
        Predicts the calibrated probability.

        Parameters
        ----------
        y_prob : 1d ndarray
            Raw probability/score of the positive class.

        Returns
        -------
        y_calibrated_prob : 1d ndarray
            Calibrated probability.
        """
        y_prob_platt = super().predict(y_prob)
        indices = np.searchsorted(self.bins_, y_prob_platt)
        return self.bins_score_[indices]


import copy
import dataclasses
import logging
import os
from typing import List, Dict, Union, Any, Optional, Literal

import keras
import numpy as np
import tensorflow as tf
import pandas as pd
from abc import abstractmethod

from keras.optimizers import Adam
#from keras_core.src.optimizers import Adam
#from keras_core.src.optimizers.schedules.learning_rate_schedule import LearningRateSchedule
#from tensorflow.keras.utils import plot_model
from keras.layers import LeakyReLU
from tensorflow.keras.utils import plot_model
# from keras.optimizers import Adam
# from keras.utils import plot_model
from keras.optimizers.schedules import LearningRateSchedule

from sklearn.utils import shuffle

from exoml.ml.callback.checkpoint import ModelCheckpointCallback
from exoml.ml.callback.learning_rate import WarmUpAndLinDecreaseCallback
from exoml.ml.callback.basemodel_aware_callback import MetricsPlotCallback
from exoml.ml.callback.batch_aware_csv_logger import BatchAwareCsvLogger
from exoml.ml.callback.early_stopping import ExoMlEarlyStopping
from exoml.ml.callback.training_data_aware import ValidationDataAwareCallback, ModelDirDataAwareCallback
from exoml.ml.learning_rate.schedulers import ExponentialRescaleDecay
from exoml.ml.log.get_weights_logger import ModelWeightsLogger
from exoml.ml.log.with_logging import WithLogging
from exoml.ml.layers.random_reverse import RandomReverseLayer
from exoml.ml.layers.dropout import SafeSpatialDropout1D
from exoml.ml.layers.branch_dropout import BranchDropoutLayer

from exoml.ml.metrics.auc import precision_at_k, mean_positive_value, mean_true_positive_value, mean_false_positive_value, \
    mean_true_negative_value, ThresholdAtPrecision

from exoml.ml.callback.learning_rate import MultiOptimizer, SWA
#from tensorflow_addons.optimizers import MultiOptimizer, SWA


@dataclasses.dataclass
class HyperParams:
    '''Number of samples to be inclded in each mini-batch'''
    batch_size: int = 20
    '''Number of iterations over the entire dataset to be run before stopping'''
    epochs: int = 20
    '''Measurement of weight differences between batches'''
    initial_learning_rate: float = 0.01,
    '''Number of iterations over the entire dataset to be done per epoch'''
    dataset_iterations_per_epoch: float = 1
    '''Percentage of samples from the entire dataset to be used as training data'''
    train_percent: float = 0.8
    '''Percentage of samples from the entire dataset to be used as validation data'''
    validation_percent: float = 0.1
    '''Number of samples to be used for training'''
    training_set_limit: Optional[int] = None
    '''Blocks the model training, allowing for all the previous steps to configure the model'''
    dry_run: bool = True
    '''Minimum value to be used instead of zero'''
    zero_epsilon = 1e-7
    '''If a gradient exceeds the threshold norm, we clip that gradient to the threshold'''
    gradient_clip_norm: Optional[float] = 0.01
    '''If a gradient exceeds the threshold value, we clip that gradient by multiplying the unit vector of the gradients 
    with the threshold'''
    gradient_clip_value: Optional[float] = 0.5
    '''Number of cores to be used'''
    cores: int = 0
    '''List of additional metrics to be computed'''
    metrics: List = dataclasses.field(default_factory=lambda: [])
    '''List of additional callbacks to be used'''
    callbacks: List = dataclasses.field(default_factory=lambda: [])
    '''Learning rate decay rate for an exponential decay'''
    learning_rate_decay: float = 0.98
    '''Custom learning rate schedule'''
    learning_rate_schedule: Optional[LearningRateSchedule] = None
    '''The class to use as reference to balance the entire dataset'''
    balance_class_id: Optional[str] = None
    '''The sampling values to be used to balance the entire dataset'''
    balance_class_sampling: Optional[List] = None
    '''The custom weights to be used to give more/less scores to some classes'''
    class_loss_weights: Optional[Union[Dict, Literal['auto']]] = None
    '''Loss difference accepted to stop the execution before the last epoch'''
    early_stopping_delta: float = 0
    '''Number of epochs to wait before stopping the execution'''
    early_stopping_patience: int = 0
    '''Custom loss function to be used'''
    custom_loss: Optional[Any] = None
    run_eagerly: bool = False
    '''L1 regularization factor'''
    l1_regularization: float = 0.0
    '''L2 regularization factor'''
    l2_regularization: float = 0.0
    '''L1 regularization factor for convolutional layers'''
    l1_regularization_conv: float = 0.0
    '''L2 regularization factor for convolutional layers'''
    l2_regularization_conv: float = 0.0
    '''Initial dropout rate'''
    branch_dropout_rate: float = 0
    '''Initial dropout rate'''
    dropout_rate: float = 0.1
    '''Maximum value for adaptive std dropout'''
    dropout_max_rate: float = 0.1
    '''Dropout for convolutional layers'''
    spatial_dropout_rate: float = 0.1
    '''Initial white noise standard deviation value'''
    white_noise_std: Optional[float] = 0.0
    '''Initial white noise standard deviation value'''
    numerical_white_noise_std: Optional[float] = 0.0
    '''Initial white noise standard deviation value for conv layers'''
    white_noise_layer_std: Optional[float] = 0.0
    '''Number of cross validation folds to be used for training'''
    cross_validation_folds: float = 0
    '''Epoch from which starting applying stochastic weight average'''
    stochastic_weight_average_wait: float = 0
    '''Learning rate progression factor to be applied to the final dense layers'''
    lr_progression: float = 1
    '''Layer normalization epsilon'''
    layer_norm_epsilon: float = 1e-3
    '''Layer normalization epsilon for transformer layers'''
    transformer_layer_norm_epsilon: float = 1e-3


@dataclasses.dataclass
class CategoricalPredictionSetStats:
    """
    Statistics of categorical predictions.
    """
    tp: int
    fp: int
    fn: int
    accuracy: float
    precision: float
    recall: float
    k_df: Optional[pd.DataFrame] = None
    predictions_df: Optional[pd.DataFrame] = None


class BaseModel(WithLogging):
    """
    Base class providing standard methods for model building.
    """

    model: keras.Model = None

    def __init__(self, name, input_size, class_ids, type_to_label, hyperparams) -> None:
        super().__init__()
        self.name = name
        self.input_size = input_size
        if class_ids is not None:
            for class_id in class_ids.keys():
                class_ids[class_id] = class_ids[class_id] if isinstance(class_ids[class_id], (list, np.ndarray)) \
                    else [class_ids[class_id]]
        self.class_ids = class_ids
        self.type_to_label = type_to_label
        self.hyperparams = hyperparams

    @abstractmethod
    def build(self, **kwargs):
        """
        Should be used to create all the model layers and related capabilities to be used in the training and subsequent
        tasks.
        :param kwargs:
        """
        pass

    @abstractmethod
    def load_training_set(self, **kwargs):
        """
        Loads the training set from its source
        :param kwargs: the case-specific params
        :return: either a list or a Pandas Dataframe
        """
        pass

    @abstractmethod
    def instance_generator(self, dataset, dir, batch_size, input_sizes, type_to_label, zero_epsilon, shuffle=True):
        pass

    @abstractmethod
    def instance_loss_accuracy(self):
        pass

    @abstractmethod
    def instance_metrics(self):
        pass

    @abstractmethod
    def plot_metrics(self, model_dir, steps_per_epoch, with_validation=False):
        pass

    def balance_dataset_with_sampling(self, training_set, classes_sampling):
        classes_counts = [len(training_set[training_set['type'].isin(labels)]) for labels in self.class_ids.values()]
        dropped_samples = training_set.iloc[:0, :].copy()
        for id, sampling in enumerate(classes_sampling):
            labels = self.class_ids[id]
            class_count = classes_counts[id]
            class_expected_size = class_count * sampling
            if sampling > 1:
                rows_for_type = training_set[training_set['type'].isin(labels)]
                sampling = sampling - 1
                training_set = [training_set, rows_for_type.sample(frac=sampling, replace=True)]
                # training_set = [training_set, rows_for_type * np.floor(sampling).astype(int)]
                # training_set = pd.concat(training_set)
                # class_expected_size_remainder = np.round(np.mod(sampling, 1) * class_count).astype(int)
                # training_set = [training_set, training_set[training_set['type'].isin(labels)].iloc[0:class_expected_size_remainder]]
                training_set = pd.concat(training_set)
            elif sampling < 1:
                class_expected_size = np.round(class_expected_size).astype(int)
                samples_to_keep = training_set[training_set['type'].isin(labels)].iloc[0:class_expected_size]
                samples_to_remove = training_set[training_set['type'].isin(labels)].iloc[class_expected_size:]
                dropped_samples = dropped_samples.append([samples_to_remove])
                training_set.drop(training_set[training_set['type'].isin(labels)].index, inplace=True)
                training_set = [training_set, samples_to_keep]
                training_set = pd.concat(training_set)
        return training_set, dropped_samples

    def balance_dataset_from_class(self, training_set, class_id):
        assert class_id in self.class_ids.keys()
        if not isinstance(training_set, pd.DataFrame):
            logging.warning("Cannot balance dataset because the input is not a dataframe")
            return training_set
        classes_counts = [len(training_set[training_set['type'].isin(labels)]) for labels in self.class_ids.values()]
        classes_sampling = [1 if id == class_id else classes_counts[class_id] / count
                            for id, count in enumerate(classes_counts)]
        return self.balance_dataset_with_sampling(training_set, classes_sampling)

    def compile(self, optimizer, loss, metrics=None, run_eagerly=False, tuner=False):
        """
        Compiles and prepares the model for training
        :param optimizer: the optimizer to be used
        :param loss: the loss to be used
        :param metrics: the metrics to be used
        :return: the object itself
        """
        if metrics is None:
            metrics = []
        metrics_str = [str(metric) for metric in metrics]
        logging.info("Compiling model with optimizer " + str(optimizer) + ", loss " + str(loss) + " and metrics [" +
                     ",".join(metrics_str) + "]")
        self.model.compile(optimizer=optimizer, loss=loss, metrics=metrics)
        self.model.run_eagerly = run_eagerly
        return self

    def inform(self, dir):
        """
        Creates summary and a visual plot of the model
        :param dir: the directory to store the plot
        :return: the object itself
        """
        logging.info("Creating model plot and summary")
        dir = dir + self._get_model_dir()
        if not os.path.exists(dir):
            os.mkdir(dir)
        try:
            plot_model(self.model, dir + '/network.png', show_shapes=True)
        except Exception as e:
            logging.exception("Can't plot model network.png")
        self.model.summary()
        ModelWeightsLogger().log_model_weights(self.model)
        return self

    def save(self, dir, model_dir=None):
        if model_dir is None:
            dest_dir = dir + '/' + self._get_model_dir()
        else:
            dest_dir = dir + '/' + model_dir
        logging.info("Saving model into %s", dest_dir)
        self.model.save(dest_dir)
        logging.info("Saved model")
        return dest_dir

    def slice_dataset(self, dataset, validation_percent=0.1):
        logging.info("Slicing dataset with validation percent = %s", validation_percent)
        dataset_length = len(dataset)
        dataset = shuffle(dataset)
        train_last_index = int(dataset_length * (1 - validation_percent))
        train_dataset = dataset[:train_last_index]
        validation_dataset = dataset[train_last_index + 1:]
        return train_dataset, validation_dataset

    def prepare_training_data(self, training_dir, output_dir, batch_size, train_percent=0.8, validation_percent=0.1,
                              training_set_limit=None, balance_class_id=None, balance_class_sampling=None):
        training_dataset = self.load_training_set(training_dir=training_dir)
        training_dataset = shuffle(training_dataset)
        training_dataset, validation_dataset = self.slice_dataset(training_dataset, validation_percent)
        test_dataset = None
        if balance_class_sampling is not None:
            training_dataset, test_dataset = \
                self.balance_dataset_with_sampling(training_dataset, balance_class_sampling)
        elif balance_class_id is not None:
            training_dataset, test_dataset = \
                self.balance_dataset_from_class(training_dataset, balance_class_id)
        training_dataset = shuffle(training_dataset)
        if training_set_limit is not None:
            if isinstance(training_dataset, pd.DataFrame):
                training_dataset = training_dataset[:training_set_limit]
            else:
                training_dataset = training_dataset[:training_set_limit]
        dataset_length = len(training_dataset)
        if self.class_ids is not None:
            for labels in self.class_ids.values():
                class_len = len(training_dataset[training_dataset['type'].isin(labels)])
                logging.info("%s (%s %%) items of class %s", class_len, class_len / len(training_dataset) * 100, labels)
        train_dataset_filename = output_dir + "/train_dataset.csv"
        validation_dataset_filename = output_dir + "/validation_dataset.csv"
        test_dataset_filename = output_dir + "/test_dataset.csv"
        logging.info("Storing train and test file names in " + train_dataset_filename + " and " +
                     validation_dataset_filename)
        test_filenames = []
        if os.path.exists(train_dataset_filename):
            os.remove(train_dataset_filename)
        if os.path.exists(validation_dataset_filename):
            os.remove(validation_dataset_filename)
        if os.path.exists(test_dataset_filename):
            os.remove(test_dataset_filename)
        if isinstance(training_dataset, pd.DataFrame):
            training_dataset.to_csv(train_dataset_filename)
            validation_dataset.to_csv(validation_dataset_filename)
            train_dataset_len = len(training_dataset)
            validation_dataset_len = len(validation_dataset)
            logging.info("Training set is of length %s (%s %%)", len(training_dataset),
                         len(training_dataset) / dataset_length * 100)
            logging.info("Validation set is of length %s (%s %%)", len(validation_dataset),
                         len(validation_dataset) / dataset_length * 100)
            logging.info("Testing set is of length %s (%s %%)", len(test_filenames),
                         len(test_filenames) / dataset_length * 100)
            if self.class_ids is not None:
                for labels in self.class_ids.values():
                    class_len = len(training_dataset[training_dataset['type'].isin(labels)])
                    logging.info("Training set contains %s (%s %%) items of class %s", class_len,
                                 class_len / train_dataset_len * 100, labels)
                for labels in self.class_ids.values():
                    class_len = len(validation_dataset[validation_dataset['type'].isin(labels)])
                    logging.info("Validation set contains %s (%s %%) items of class %s", class_len,
                                 class_len / validation_dataset_len * 100, labels)
        return training_dataset, validation_dataset, test_filenames

    def prepare_training_data_cv(self, training_dir, folds=10):
        dataset = self.load_training_set(training_dir=training_dir)
        dataset = shuffle(dataset)
        dataset.reset_index(inplace=True, drop=True)
        dataset_len = len(dataset)
        for labels in self.class_ids.values():
            class_len = len(dataset[dataset['type'].isin(labels)])
            logging.info("%s (%s %%) items of class %s", class_len, class_len / len(dataset) * 100, labels)
        fold_indexes = [int(fold_index) for fold_index in np.linspace(dataset_len // folds, dataset_len, folds)]
        return dataset, fold_indexes

    def compute_class_weights(self, train_filenames):
        train_len = len(train_filenames)
        label_counts = {}
        for type, label in self.type_to_label.items():
            class_len = len(train_filenames[train_filenames['type'] == type])
            if label[0] in label_counts:
                label_counts[label[0]] = label_counts[label[0]] + class_len
            else:
                label_counts[label[0]] = class_len
        class_weights = {}
        for label, count in label_counts.items():
            logging.info("Label %s contains %s (%s %%) items", label, count,
                         count / train_len * 100)
            class_weights[label] = 1 / (count / train_len)
        return class_weights

    def train(self, training_dir, output_dir, hyperparams, plot_metrics=False, continue_from_model=None):
        logging.info("Preparing training data with (training_dir," + str(training_dir) +
                     ") (train_percent," + str(hyperparams.train_percent) +
                     ") (test_percent," + str(hyperparams.validation_percent) +
                     ") (training_set_limit," + str(hyperparams.training_set_limit) +
                     ")")
        model_path = output_dir + self._get_model_dir()
        if not os.path.exists(model_path):
            os.mkdir(model_path)
        train_filenames, validation_filenames, test_filenames = \
            self.prepare_training_data(training_dir, model_path, hyperparams.batch_size, hyperparams.train_percent,
                                       hyperparams.validation_percent, hyperparams.training_set_limit,
                                       hyperparams.balance_class_id, hyperparams.balance_class_sampling)
        # The optimizer is executed once for every batch, hence optimizer steps per epoch are
        train_dataset_size = len(train_filenames)
        test_dataset_size = len(validation_filenames)
        steps_per_epoch = int(hyperparams.dataset_iterations_per_epoch * train_dataset_size // hyperparams.batch_size)
        total_steps = steps_per_epoch * hyperparams.epochs
        logging.info("Initializing optimizer with (initial_learning_rate," + str(hyperparams.initial_learning_rate) +
                     ")")
        optimizer = self.build_optimizer(hyperparams, steps_per_epoch)
        # We don't use SparseCategoricalCrossentropy because our targets are one-hot encoded
        loss, accuracy = self.instance_loss_accuracy()
        if continue_from_model is not None:
            logging.info(f"Continuing model fit from checkpoint {continue_from_model}")
            self.load_model(continue_from_model)
        metrics = [accuracy] if accuracy is not None else []
        metrics = metrics + self.instance_metrics() + hyperparams.metrics
        self.compile(optimizer, hyperparams.custom_loss if hyperparams.custom_loss is not None else loss,
                     metrics=metrics,
                     run_eagerly=hyperparams.run_eagerly)
        model_weights_logger = ModelWeightsLogger()
        # if hyperparams.learning_rate_schedule is None:
        #     learning_rate_decay_steps = steps_per_epoch // 2
        #     learning_rate_decay_steps = learning_rate_decay_steps if learning_rate_decay_steps > steps_per_epoch \
        #         else steps_per_epoch
        #     logging.info("Initializing optimizer with learning_rate_decay_steps," + str(learning_rate_decay_steps) +
        #                  ") (gradient_clip_norm," + str(hyperparams.gradient_clip_norm) +
        #                  ")")
        #     powers_for_half_learning_rate_decay = np.log(hyperparams.learning_rate_decay / 2) // \
        #                                           np.log(hyperparams.learning_rate_decay)
        #     hyperparams.learning_rate_schedule = ExponentialRescaleDecay(hyperparams.initial_learning_rate,
        #                                                      decay_steps=learning_rate_decay_steps,
        #                                                      decay_rate=hyperparams.learning_rate_decay,
        #                                                      restore_steps=learning_rate_decay_steps *
        #                                                                    powers_for_half_learning_rate_decay,
        #                                                      restore_rate=1.5,
        #                                                      staircase=True)

        if not hyperparams.dry_run:
            logging.info("Initializing training and validation generators with (batch_size," + str(hyperparams.batch_size) +
                         ") (self.input_size," + str(self.input_size) +
                         ") (zero_epsilon," + str(hyperparams.zero_epsilon) +
                         ")")
            training_batch_generator = self.instance_generator(train_filenames, training_dir, hyperparams.batch_size,
                                                               self.input_size, self.type_to_label, hyperparams.zero_epsilon)
            validation_batch_generator = self.instance_generator(validation_filenames, training_dir, hyperparams.batch_size,
                                                                 self.input_size, self.type_to_label, hyperparams.zero_epsilon,
                                                                 shuffle=False)
            for callback in hyperparams.callbacks:
                if issubclass(callback.__class__, ModelDirDataAwareCallback):
                    callback.set_model_dir(model_path)
                if issubclass(callback.__class__, ValidationDataAwareCallback):
                    callback.set_validation_data(validation_batch_generator)
            additional_callbacks = [BatchAwareCsvLogger(model_path + '/training_log.csv', steps_per_epoch)]
            if plot_metrics:
                additional_callbacks = additional_callbacks + [MetricsPlotCallback(self, model_path, steps_per_epoch)]
            hyperparams.callbacks = additional_callbacks + hyperparams.callbacks
            if hyperparams.early_stopping_patience > 0 and hyperparams.early_stopping_delta > 0:
                hyperparams.callbacks = hyperparams.callbacks + [ExoMlEarlyStopping(
                    monitor="val_loss",
                    min_delta=hyperparams.early_stopping_delta,
                    patience=hyperparams.early_stopping_patience,
                    verbose=0,
                    mode="auto",
                    baseline=None,
                    restore_best_weights=False,
                )]
            model_validation_steps = None
            class_weights = hyperparams.class_loss_weights if hyperparams.class_loss_weights is not None \
                else training_batch_generator.class_weights()
            logging.info("Class weights are %s", class_weights)
            if isinstance(class_weights, str) and 'auto' == class_weights:
                class_weights = self.compute_class_weights(train_filenames)
                logging.info("Class weights are %s", class_weights)
            logging.info("Initializing training with (epochs," + str(hyperparams.epochs) +
                         ") (steps_per_epoch," + str(steps_per_epoch) +
                         ")")
            self.save(output_dir, self._get_model_prefix() + '.h5')
            self.__write_hyperparameters(hyperparams, model_path)
            fit_history = self.fit_model(hyperparams, training_batch_generator, steps_per_epoch, class_weights,
                                         validation_batch_generator, model_validation_steps)
            self.save(output_dir, self._get_model_prefix() + '.h5')
        else:
            logging.warning("dry_run was activated and 'training' will not be launched")

    def fit_model(self, hyperparams, training_batch_generator, steps_per_epoch, class_weights,
                  validation_batch_generator, model_validation_steps):
        fit_history = self.model.fit(x=training_batch_generator,
                                     steps_per_epoch=steps_per_epoch,
                                     epochs=hyperparams.epochs, verbose=1, class_weight=class_weights,
                                     validation_data=validation_batch_generator,
                                     validation_steps=model_validation_steps,
                                     callbacks=hyperparams.callbacks)
        return fit_history

    def build_optimizer(self, hyperparams, steps_per_epoch):
        initial_lr = self.compute_initial_lr(hyperparams)
        optimizer = self.build_swa_optimizer(hyperparams, initial_lr, steps_per_epoch)
        if hyperparams.lr_progression != 1:
            optimizer.progressive_lr_factor = 1
            optimizers_and_layers = []
            standard_layers = []
            progressive_lr_factor = 1
            for layer in self.model.layers:
                if 'final' in layer.name:
                    progressive_lr_factor = progressive_lr_factor * hyperparams.lr_progression
                    progressive_optimizer = self.build_swa_optimizer(hyperparams, progressive_lr_factor * initial_lr, steps_per_epoch)
                    progressive_optimizer.progressive_lr_factor = progressive_lr_factor
                    optimizers_and_layers = optimizers_and_layers + [
                        (progressive_optimizer, layer)]
                else:
                    standard_layers = standard_layers + [layer]
            optimizers_and_layers = [(optimizer, standard_layers)] + optimizers_and_layers
            optimizer = MultiOptimizer(optimizers_and_layers)
        return optimizer

    def compute_initial_lr(self, hyperparams: HyperParams):
        if hyperparams.lr_progression != 1:
            initial_lr = hyperparams.initial_learning_rate
        else:
            initial_lr = hyperparams.learning_rate_schedule \
                if hyperparams.learning_rate_schedule is not None else hyperparams.initial_learning_rate
        return initial_lr

    def build_swa_optimizer(self, hyperparams: HyperParams, lr, steps_per_epoch):
        if hyperparams.stochastic_weight_average_wait > 0:
            optimizer = tf.keras.optimizers.legacy.Adam(lr,
                                                        beta_1=0.9, beta_2=0.98, epsilon=1e-9,
                                                        clipnorm=hyperparams.gradient_clip_norm,
                                                        clipvalue=hyperparams.gradient_clip_value)
            optimizer = SWA(optimizer,
                                           start_averaging=steps_per_epoch * hyperparams.stochastic_weight_average_wait,
                                           average_period=steps_per_epoch)
            if isinstance(lr, (int, float)):
                optimizer.lr = lr
            else:
                optimizer.lr = hyperparams.initial_learning_rate
        else:
            optimizer = Adam(lr,
                                                 beta_1=0.9, beta_2=0.98, epsilon=1e-9,
                                                 clipnorm=hyperparams.gradient_clip_norm,
                                                 clipvalue=hyperparams.gradient_clip_value)
        return optimizer

    @staticmethod
    def log_sets_distribution(training_set, validation_set, dataset_length, class_ids, testing_set=None):
        train_dataset_size = len(training_set)
        validation_dataset_size = len(validation_set)
        logging.info("Training set is of length %s (%s %%)", train_dataset_size,
                     train_dataset_size / dataset_length * 100)
        logging.info("Validation set is of length %s (%s %%)", validation_dataset_size,
                     validation_dataset_size / dataset_length * 100)
        for labels in class_ids.values():
            class_len = len(training_set[training_set['type'].isin(labels)])
            logging.info("Training set contains %s (%s %%) items of class %s", class_len,
                         class_len / train_dataset_size * 100, labels)
        for labels in class_ids.values():
            class_len = len(validation_set[validation_set['type'].isin(labels)])
            logging.info("Validation set contains %s (%s %%) items of class %s", class_len,
                         class_len / validation_dataset_size * 100, labels)

    def train_cv(self, training_dir, output_dir, hyperparams, folds=10, retry_indexes=[],
                 continue_from=None, plot_metrics=False, devices=None):
        root_model_name = self._get_model_dir()
        model_name_initial = root_model_name.strip("/") + '_initial/'
        dataset, fold_indexes = self.prepare_training_data_cv(training_dir, folds)
        dataset_length = len(dataset)
        logging.info("Preparing training data with (training_dir," + str(training_dir) +
                     ") (train_percent," + str(hyperparams.train_percent) +
                     ") (test_percent," + str(hyperparams.validation_percent) +
                     ") and CV Folds with indexes %s", fold_indexes)
        # We don't use SparseCategoricalCrossentropy because our targets are one-hot encoded
        loss, accuracy = self.instance_loss_accuracy()
        metrics = [accuracy] if accuracy is not None else []
        metrics = metrics + self.instance_metrics() + hyperparams.metrics
        for index, fold_index in enumerate(fold_indexes):
            model_name = root_model_name.strip("/") + f'_{index}/'
            model_path = output_dir + '/' + model_name
            if not os.path.exists(model_path):
                os.mkdir(model_path)
                previous_fold_index = fold_indexes[index - 1] if index > 0 else 0
                training_set = pd.concat([dataset.iloc[0:previous_fold_index], dataset.iloc[fold_index:]])
                validation_set = dataset.iloc[previous_fold_index:fold_index]
                logging.info("Generating training for K-fold no %s", index)
                BaseModel.log_sets_distribution(training_set, validation_set, dataset_length, self.class_ids)
                training_set.to_csv(model_path + 'train_dataset.csv')
                validation_set.to_csv(model_path + 'validation_dataset.csv')
        optimizer = self.build_optimizer(hyperparams, 150)
        self.compile(optimizer, hyperparams.custom_loss if hyperparams.custom_loss is not None else loss,
                     metrics=metrics,
                     run_eagerly=hyperparams.run_eagerly)
        self.save(output_dir, model_name_initial)
        for index, fold_index in enumerate(fold_indexes):
            with self.train_strategy(devices).scope():
                model_name = root_model_name.strip("/") + f'_{index}/'
                model_chk_name = root_model_name.strip("/") + f'_chk_{index}/'
                model_path = output_dir + '/' + model_name
                model_chk_path = output_dir + '/' + model_chk_name
                if not os.path.exists(model_path):
                    os.mkdir(model_path)
                original_checkpoint_callback = None
                if len(retry_indexes) == 0 or (len(retry_indexes) > 0 and index in retry_indexes):
                    logging.info("Preparing training for K-fold no %s", index)
                    training_set = pd.read_csv(model_path + '/train_dataset.csv')
                    validation_set = pd.read_csv(model_path + '/validation_dataset.csv')
                    train_dataset_size = len(training_set)
                    validation_dataset_size = len(validation_set)
                    BaseModel.log_sets_distribution(training_set, validation_set, dataset_length, self.class_ids)
                    logging.info("Initializing optimizer with (initial_learning_rate," +
                                 str(hyperparams.initial_learning_rate) + ")")
                    steps_per_epoch = int(hyperparams.dataset_iterations_per_epoch * train_dataset_size // hyperparams.batch_size)
                    optimizer = self.build_optimizer(hyperparams, steps_per_epoch)
                    if not hyperparams.dry_run:
                        logging.info("Initializing training and validation generators with (batch_size," + str(hyperparams.batch_size) +
                                     ") (self.input_size," + str(self.input_size) +
                                     ") (zero_epsilon," + str(hyperparams.zero_epsilon) +
                                     ")")
                        training_batch_generator = self.instance_generator(training_set, training_dir, hyperparams.batch_size,
                                                                           self.input_size, self.type_to_label, hyperparams.zero_epsilon)
                        validation_batch_generator = self.instance_generator(validation_set, training_dir, hyperparams.batch_size,
                                                                             self.input_size, self.type_to_label, hyperparams.zero_epsilon,
                                                                             shuffle=False)
                        callbacks = []
                        for callback in hyperparams.callbacks:
                            new_callback = callback
                            logging.info(f"callback class {callback.__class__.__name__}")
                            if callback.__class__.__name__ == ModelDirDataAwareCallback.__name__:
                                new_callback.set_model_dir(model_name)
                            if callback.__class__.__name__ == ValidationDataAwareCallback.__name__:
                                new_callback.set_validation_data(validation_batch_generator)
                            if callback.__class__.__name__ == ModelCheckpointCallback.__name__:
                                if original_checkpoint_callback is None:
                                    original_checkpoint_callback = copy.deepcopy(callback)
                                    logging.info(f"Stored original checkpoint callback with "
                                                 f"{original_checkpoint_callback.original_path} "
                                                 f",{original_checkpoint_callback.filepath} and "
                                                 f"{original_checkpoint_callback.best}")
                                new_callback = copy.deepcopy(original_checkpoint_callback)
                                logging.info(f"Switched modelCheckpointCallback to {model_path} "
                                             f",{model_chk_path} and {new_callback.best}")
                                new_callback.original_path = model_path
                                new_callback.filepath = model_chk_path + '/' + os.path.basename(model_chk_path) + '.keras'
                            callbacks = callbacks + [new_callback]
                        additional_callbacks = [BatchAwareCsvLogger(model_name + '/training_log.csv', steps_per_epoch)]
                        if plot_metrics:
                            additional_callbacks = additional_callbacks + [
                                MetricsPlotCallback(self, model_name, steps_per_epoch)]
                        hyperparams.callbacks = additional_callbacks + hyperparams.callbacks
                        callbacks = additional_callbacks + callbacks
                        if hyperparams.early_stopping_patience > 0 and hyperparams.early_stopping_delta > 0:
                            callbacks = callbacks + [ExoMlEarlyStopping(
                                monitor="val_loss",
                                min_delta=hyperparams.early_stopping_delta,
                                patience=hyperparams.early_stopping_patience,
                                verbose=0,
                                mode="auto",
                                baseline=None,
                                restore_best_weights=False,
                            )]
                        model_validation_steps = int(validation_dataset_size // hyperparams.batch_size)
                        class_weights = hyperparams.class_loss_weights if hyperparams.class_loss_weights is not None \
                            else training_batch_generator.class_weights()
                        if isinstance(class_weights, str) and 'auto' == class_weights:
                            class_weights = self.compute_class_weights(training_set)
                        logging.info("Initializing training with (epochs," + str(hyperparams.epochs) +
                                     ") (steps_per_epoch," + str(steps_per_epoch) +
                                     ") (model_validation_steps," + str(model_validation_steps) +
                                     ")")
                        continue_from_model = continue_from + str(index)
                        if continue_from is not None and os.path.exists(continue_from_model):
                            logging.info(f"Continuing model fit from checkpoint {continue_from_model}")
                            self.load_model(continue_from_model)
                        else:
                            logging.info(f"Continuing model fit from initial {model_name_initial}")
                            self.load_model(model_name_initial)
                        self.compile(optimizer,
                                     hyperparams.custom_loss if hyperparams.custom_loss is not None else loss,
                                     metrics=metrics,
                                     run_eagerly=hyperparams.run_eagerly)
                        self.__write_hyperparameters(hyperparams, model_name)
                        fit_history = self.model.fit(x=training_batch_generator,
                                       steps_per_epoch=steps_per_epoch,
                                       epochs=hyperparams.epochs, verbose=1, class_weight=class_weights,
                                       validation_data=validation_batch_generator,
                                       validation_steps=model_validation_steps,
                                       callbacks=callbacks,
                                       use_multiprocessing=hyperparams.cores > 0, workers=1 if hyperparams.cores <= 0
                            else hyperparams.cores)
                        self.save(output_dir, model_name)
                    else:
                        logging.warning("dry_run was activated and 'training' will not be launched")

    def fine_tune_cv(self, training_dir, output_dir, hyperparams, continue_from, folds=10, retry_indexes=[],
                 plot_metrics=False, devices=None, regenerate_fine_tuning_set=True, validate_only_ft=False,
                     train_only_ft=True):
        root_model_name = self._get_model_dir()
        # We don't use SparseCategoricalCrossentropy because our targets are one-hot encoded
        loss, accuracy = self.instance_loss_accuracy()
        metrics = [accuracy] if accuracy is not None else []
        metrics = [accuracy] + self.instance_metrics() + hyperparams.metrics
        if regenerate_fine_tuning_set:
            dataset, fold_indexes = self.prepare_training_data_cv(training_dir, folds)
            dataset_length = len(dataset)
            logging.info("Preparing fine tuning data with (training_dir," + str(training_dir) +
                         ") (train_percent," + str(hyperparams.train_percent) +
                         ") (test_percent," + str(hyperparams.validation_percent) +
                         ") and CV Folds with indexes %s", fold_indexes)
            for index, fold_index in enumerate(fold_indexes):
                model_name = root_model_name.strip("/") + f'_{index}/'
                model_path = output_dir + '/' + model_name
                previous_fold_index = fold_indexes[index - 1] if index > 0 else 0
                training_set = pd.concat([dataset.iloc[0:previous_fold_index], dataset.iloc[fold_index:]])
                validation_set = dataset.iloc[previous_fold_index:fold_index]
                logging.info("Generating fine-tuning for K-fold no %s", index)
                BaseModel.log_sets_distribution(training_set, validation_set, dataset_length, self.class_ids)
                training_set.to_csv(model_path + 'ft_train_dataset.csv')
                validation_set.to_csv(model_path + 'ft_validation_dataset.csv')
        for index in np.arange(folds):
            with self.train_strategy(devices).scope():
                model_name = root_model_name.strip("/") + f'_{index}/'
                model_chk_name = root_model_name.strip("/") + f'_chk_{index}/'
                model_path = output_dir + '/' + model_name
                model_chk_path = output_dir + '/' + model_chk_name
                original_checkpoint_callback = None
                if len(retry_indexes) == 0 or (len(retry_indexes) > 0 and index in retry_indexes):
                    logging.info("Preparing training for K-fold no %s", index)
                    training_set = pd.read_csv(model_path + '/ft_train_dataset.csv')
                    if not train_only_ft:
                        training_set = pd.concat([training_set, pd.read_csv(model_path + '/train_dataset.csv')])
                    validation_set = pd.read_csv(model_path + '/ft_validation_dataset.csv')
                    if not validate_only_ft:
                        validation_set = pd.concat([validation_set, pd.read_csv(model_path + '/validation_dataset.csv')])
                    train_dataset_size = len(training_set)
                    validation_dataset_size = len(validation_set)
                    BaseModel.log_sets_distribution(training_set, validation_set, train_dataset_size + train_dataset_size, self.class_ids)
                    logging.info("Initializing optimizer with (initial_learning_rate," +
                                 str(hyperparams.initial_learning_rate) + ")")
                    steps_per_epoch = int(hyperparams.dataset_iterations_per_epoch * train_dataset_size // hyperparams.batch_size)
                    optimizer = self.build_optimizer(hyperparams, steps_per_epoch)
                    if not hyperparams.dry_run:
                        logging.info("Initializing training and validation generators with (batch_size," + str(hyperparams.batch_size) +
                                     ") (self.input_size," + str(self.input_size) +
                                     ") (zero_epsilon," + str(hyperparams.zero_epsilon) +
                                     ")")
                        training_batch_generator = self.instance_generator(training_set, training_dir, hyperparams.batch_size,
                                                                           self.input_size, self.type_to_label, hyperparams.zero_epsilon)
                        validation_batch_generator = self.instance_generator(validation_set, training_dir, hyperparams.batch_size,
                                                                             self.input_size, self.type_to_label, hyperparams.zero_epsilon,
                                                                             shuffle=False)
                        callbacks = []
                        for callback in hyperparams.callbacks:
                            new_callback = callback
                            logging.info(f"callback class {callback.__class__.__name__}")
                            if callback.__class__.__name__ == ModelDirDataAwareCallback.__name__:
                                new_callback.set_model_dir(model_name)
                            if callback.__class__.__name__ == ValidationDataAwareCallback.__name__:
                                new_callback.set_validation_data(validation_batch_generator)
                            if callback.__class__.__name__ == ModelCheckpointCallback.__name__:
                                if original_checkpoint_callback is None:
                                    original_checkpoint_callback = copy.deepcopy(callback)
                                    logging.info(f"Stored original checkpoint callback with "
                                                 f"{original_checkpoint_callback.original_path} "
                                                 f",{original_checkpoint_callback.filepath} and "
                                                 f"{original_checkpoint_callback.best}")
                                new_callback = copy.deepcopy(original_checkpoint_callback)
                                logging.info(f"Switched modelCheckpointCallback to {model_path} "
                                             f",{model_chk_path} and {new_callback.best}")
                                new_callback.original_path = model_path
                                new_callback.filepath = model_chk_path
                            callbacks = callbacks + [new_callback]
                        additional_callbacks = [BatchAwareCsvLogger(model_name + '/training_log.csv', steps_per_epoch)]
                        if plot_metrics:
                            additional_callbacks = additional_callbacks + [
                                MetricsPlotCallback(self, model_name, steps_per_epoch)]
                        hyperparams.callbacks = additional_callbacks + hyperparams.callbacks
                        callbacks = additional_callbacks + callbacks
                        if hyperparams.early_stopping_patience > 0 and hyperparams.early_stopping_delta > 0:
                            callbacks = callbacks + [ExoMlEarlyStopping(
                                monitor="val_loss",
                                min_delta=hyperparams.early_stopping_delta,
                                patience=hyperparams.early_stopping_patience,
                                verbose=0,
                                mode="auto",
                                baseline=None,
                                restore_best_weights=False,
                            )]
                        model_validation_steps = int(validation_dataset_size // hyperparams.batch_size)
                        class_weights = hyperparams.class_loss_weights if hyperparams.class_loss_weights is not None \
                            else training_batch_generator.class_weights()
                        if isinstance(class_weights, str) and 'auto' == class_weights:
                            class_weights = self.compute_class_weights(training_set)
                        logging.info("Initializing training with (epochs," + str(hyperparams.epochs) +
                                     ") (steps_per_epoch," + str(steps_per_epoch) +
                                     ") (model_validation_steps," + str(model_validation_steps) +
                                     ")")
                        continue_from_model = continue_from + str(index)
                        logging.info(f"Continuing model fit from checkpoint {continue_from_model}")
                        self.load_model(continue_from_model)
                        # for layer in self.model.layers:
                        #     if not isinstance(layer, (keras.layers.GaussianNoise, keras.layers.InputLayer, keras.layers.Dense)):
                        #         layer.trainable = False
                        self.compile(optimizer,
                                     hyperparams.custom_loss if hyperparams.custom_loss is not None else loss,
                                     metrics=metrics,
                                     run_eagerly=hyperparams.run_eagerly)
                        self.__write_hyperparameters(hyperparams, model_name)
                        fit_history = self.model.fit(x=training_batch_generator,
                                       steps_per_epoch=steps_per_epoch,
                                       epochs=hyperparams.epochs, verbose=1, class_weight=class_weights,
                                       validation_data=validation_batch_generator,
                                       validation_steps=model_validation_steps,
                                       callbacks=callbacks,
                                       use_multiprocessing=hyperparams.cores > 0, workers=1 if hyperparams.cores <= 0
                            else hyperparams.cores)
                        self.save(output_dir, model_name)
                    else:
                        logging.warning("dry_run was activated and 'training' will not be launched")

    def train_strategy(self, devices=None):
        gpus = tf.config.list_physical_devices('GPU')
        num_gpus = len(gpus)
        logging.info("Number of GPUs available: %s", num_gpus)
        for gpu in gpus:
            logging.info(gpu)
        # Use MirroredStrategy if more than one GPU is available
        if devices is not None and len(devices) > 1:
            logging.info("Using MirroredStrategy for multi-GPU training.")
            strategy = tf.distribute.MirroredStrategy(devices)
        else:
            logging.info("Using default strategy for single GPU or CPU.")
            if devices is not None:
                logging.info("Selecting device %s", devices)
                tf.config.set_visible_devices([tf.config.PhysicalDevice(name=device, device_type='GPU') for device in devices], 'GPU')
            strategy = tf.distribute.get_strategy()  # Default strategy that works on a single GPU or CPU
        return strategy

    def _apply_normalization_mode(self, layer, normalization_mode='layer_norm'):
        if normalization_mode == 'batch_norm':
            final_layer = keras.layers.BatchNormalization()(layer)
        elif normalization_mode == 'layer_norm':
            final_layer = keras.layers.LayerNormalization()(layer)
        else:
            final_layer = layer
        return final_layer


    def __write_hyperparameters(self, hyperparameters: HyperParams, output_dir):
        pass
        #only_attributes_hyperparameters = deepcopy(hyperparameters)
        # yaml = ruamel.yaml.YAML()
        # yaml.register_class(HyperParams)
        # # if len(hyperparameters.callbacks) > 0:
        # #     hyperparameters.callbacks = []
        # #     for i, callback in enumerate(hyperparameters.callbacks):
        # #         only_attributes_hyperparameters.callbacks = only_attributes_hyperparameters.callbacks + \
        # #                                                     [str(type(hyperparameters.callbacks[i]))]
        # # if only_attributes_hyperparameters.learning_rate_schedule is not None:
        # #     only_attributes_hyperparameters.learning_rate_schedule = json.dumps(
        # #         dataclasses.asdict(only_attributes_hyperparameters.learning_rate_schedule))
        # # if only_attributes_hyperparameters.custom_loss is not None:
        # #     only_attributes_hyperparameters.custom_loss = str(type(hyperparameters.custom_loss))
        # with open(output_dir + '/hp.yaml', 'w', newline='') as f:
        #     yaml.dump(hyperparameters, f)

    def load_model(self, dir, custom_objects={}, compile=False):
        custom_objects['BranchDropoutLayer'] = BranchDropoutLayer
        custom_objects['SpatialDropout1D'] = SafeSpatialDropout1D
        custom_objects['LeakyReLU'] = LeakyReLU
        custom_objects['RandomReverseLayer'] = RandomReverseLayer
        custom_objects['ExponentialRescaleDecay'] = ExponentialRescaleDecay
        custom_objects['WarmUpAndLinDecreaseCallback'] = WarmUpAndLinDecreaseCallback
        custom_objects['precision_at_k'] = precision_at_k
        custom_objects['mean_false_positive_value'] = mean_false_positive_value
        custom_objects['mean_true_positive_value'] = mean_true_positive_value
        custom_objects['mean_true_negative_value'] = mean_true_negative_value
        custom_objects['mean_false_negative_value'] = mean_false_positive_value
        custom_objects['ThresholdAtPrecision'] = None
        logging.info("Loading model from %s", dir)
        self.set_model(tf.keras.models.load_model(dir, compile=compile,
                                                  custom_objects=custom_objects))
        return self

    def set_model(self, model: keras.Model):
        """
        Stores the model in a class attribute
        :param model: the model to be stored
        """
        self.model = model

    def _get_model_dir(self):
        return self._get_model_prefix() + '/'

    def _get_model_prefix(self):
        return self.name + '_model'




import logging
import threading

import numpy as np
import pandas as pd
from sklearn.utils import shuffle
from tensorflow.keras.utils import Sequence
from matplotlib import pyplot as plt
from functools import lru_cache


@lru_cache(maxsize=5)  # keep up to 5 files cached
def load_flux(path, delimiter=','):
    return np.loadtxt(path, delimiter=delimiter)


class SantoGenerator(Sequence):
    def __init__(
            self,
            lcs_dir,
            target_files,
            input_size=500,
            step_size=1,
            batch_size=500,
            shuffle=True,
            zero_epsilon=1e-7,
            indexes_steps=1,
            plot=False,
            from_arrays=False,
            store_arrays=False
    ):
        self.lcs_dir = lcs_dir
        self.target_files = target_files
        self.input_size = input_size
        self.step_size = step_size
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.zero_epsilon = zero_epsilon
        self.plot = plot
        self.indexes_steps = indexes_steps
        self.from_arrays = from_arrays
        self.store_arrays = store_arrays
        self.steps_count, self.generator_info_df = self._count_inputs()

    def _count_inputs(self):
        batch_index = 0
        file_count = 0

        # Create a list to collect all rows and build DataFrame once at the end
        data_rows = []

        for target_file in self.target_files:
            if not (target_file.endswith("cadences.csv") or target_file.endswith("flux.csv") or target_file.endswith("tags.csv")):
                # Load file dimensions only once per file
                flux = load_flux(f"{self.lcs_dir}/{target_file}", delimiter=",")

                # Calculate all batch indices at once
                #file_batch_indexes = np.arange(
                #    -self.input_size // 2,
                #    flux.shape[1] - self.input_size // 2 - self.batch_size - 1,
                #    self.batch_size,
                #)
                file_batch_indexes = np.arange(
                    0,
                    flux.shape[1] - self.input_size - self.batch_size - 1,
                    self.batch_size,
                )

                logging.info(f"Doing file {file_count} with {len(file_batch_indexes)} batches")

                # Add all rows for this file to our list at once
                for file_index in file_batch_indexes:
                    data_rows.append({
                        "filename": target_file,
                        "file_index": file_index,
                        "batch_index": batch_index
                    })
                    batch_index += 1

                file_count += 1

        # Create DataFrame once with all data
        generator_info_df = pd.DataFrame(data_rows) if data_rows else pd.DataFrame(
            columns=["filename", "file_index", "batch_index"])

        return (batch_index - 1) if batch_index > 0 else 0, generator_info_df

    def __len__(self):
        return self.generator_info_df["batch_index"].max()

    def save_arrays(self, filename, iteration_index, flux_data, cadences, tags_data):
        np.savetxt(f"{self.lcs_dir}/{filename}_{iteration_index}_flux.csv", flux_data[0], delimiter=',')
        np.savetxt(f"{self.lcs_dir}/{filename}_{iteration_index}_cadences.csv", [cadences], delimiter=',')
        np.savetxt(f"{self.lcs_dir}/{filename}_{iteration_index}_tags.csv", tags_data, delimiter=',')

    def __getitem__(self, index):
        filename: str = self.generator_info_df.iloc[index]["filename"]
        file_index: int = self.generator_info_df.loc[index]["file_index"]
        flux = load_flux(f"{self.lcs_dir}/{filename}", delimiter=",")
        # train_fluxes = np.full((self.batch_size, self.input_size, 33), 0)
        train_fluxes = np.full((self.batch_size, self.input_size), 0.0)
        train_tags = np.full((self.batch_size, self.input_size), float(0))
        cadences = np.full((self.batch_size), 0.0)
        for iteration_index in np.arange(file_index, file_index + self.batch_size):
            item_index = iteration_index - file_index
            try:
                read = False
                if self.from_arrays:
                    try:
                        train_fluxes[item_index] = np.loadtxt(f"{self.lcs_dir}/{filename}_{iteration_index}_flux.csv",  delimiter=',')
                        train_tags[item_index] = np.loadtxt(f"{self.lcs_dir}/{filename}_{iteration_index}_tags.csv", delimiter=',')
                        cadences[item_index] = np.loadtxt(f"{self.lcs_dir}/{filename}_{iteration_index}_cadences.csv", delimiter=',')
                        read = True
                    except:
                        read = False
                if not self.from_arrays or not read:
                    #if iteration_index < 0:
                        # flux_data = flux[1, 0 : iteration_index + self.input_size]
                        # flux_data = np.pad(flux_data, [(self.input_size - flux_data.shape[0], 0)], mode='constant', constant_values=0)
                    #    flux_data = flux[np.r_[1, 0], 0:iteration_index + self.input_size]
                    #    tags_data = flux[3, 0 : iteration_index + self.input_size]
                        #flux_data = np.pad(flux_data, [(0, 0), (self.input_size - flux_data.shape[1], 0)],
                        #                   mode='constant', constant_values=0)
                        #tags_data = np.pad(tags_data, [(self.input_size - tags_data.shape[0], 0)],
                        #                   mode='constant', constant_values=0)
                    #elif iteration_index >= flux.shape[1] - self.input_size:
                        # flux_data = flux[1, iteration_index:flux.shape[1]]
                        # flux_data = np.pad(flux_data, [(0, self.input_size - flux_data.shape[0])], mode='constant', constant_values=0)
                    #    flux_data = flux[np.r_[1, 0], iteration_index:flux.shape[1]]
                    #    tags_data = flux[3, iteration_index : flux.shape[1]]
                        #flux_data = np.pad(flux_data, [(0, 0), (0, self.input_size - flux_data.shape[1])],
                        #                   mode='constant', constant_values=0)
                        #tags_data = np.pad(tags_data, [(0, self.input_size - tags_data.shape[0])],
                        #                   mode='constant', constant_values=0)
                    #else:
                    flux_data = flux[np.r_[1, 0], iteration_index:iteration_index + self.input_size]
                    tags_data = flux[3, iteration_index: iteration_index + self.input_size]
                        # flux_data = flux[1, iteration_index:iteration_index + self.input_size]
                    flux_data[1][1:] = np.where(
                        flux_data[1][1:] != 0,
                        np.minimum(flux_data[1][1:] - flux_data[1][:-1], 1 - self.zero_epsilon),
                        flux_data[1][1:]
                    )
                    flux_data[1][0] = 0
                    # tags_data = flux[3, iteration_index + self.input_size // 2]

                    # import matplotlib.pyplot as plt
                    #
                    # fig, axs = plt.subplots(2, 1, figsize=(16, 16), constrained_layout=True)
                    # axs[0].scatter(np.arange(0, len(np.transpose(flux_data[0].flatten()))),
                    #                np.transpose(flux_data[0].flatten()))
                    # axs[1].plot(np.arange(0, len(np.transpose(flux_data[0].flatten()))),
                    #             np.transpose(flux[3, iteration_index:iteration_index + self.input_size]).flatten())
                    # plt.show()
                    flux_data[0] = flux_data[0] / 2
                    # flux_data = flux_data / 2
                    # train_fluxes[item_index] = flux_data.reshape((self.input_size, 1))
                    #train_fluxes[item_index] = np.transpose(flux_data)
                    train_fluxes[item_index] = flux_data[0]
                    train_tags[item_index] = tags_data
                    cadences[item_index] = np.median(flux_data[1][np.argwhere(flux_data[1] > 0).flatten()]) * 24
                    cadences[item_index] = np.where(cadences[item_index] >= 1, 1 - self.zero_epsilon, cadences[item_index])
                    if self.store_arrays or (not read and self.from_arrays):
                        threading.Thread(
                            target=self.save_arrays,
                            args=(filename, iteration_index, flux_data, cadences[item_index], tags_data),
                            daemon=True
                        ).start()
                        #logging.info(f"Storing arrays for {filename} iteration {iteration_index}")
            except Exception as e:
                logging.exception(
                    f"Error with file {filename}. file_index {file_index}, "
                    f"iteration_index {iteration_index}, batch_size {self.batch_size}, "
                    f"input_size {self.input_size}"
                )
                raise e
            # if self.plot:
            #     fig, ax = plt.subplots(2, 1, figsize=(16, 8), constrained_layout=True)
            #     ax[0].scatter(np.arange(0, len(train_fluxes[item_index])), train_fluxes[item_index])
            #     ax[1].plot(np.arange(0, train_tags[item_index]), train_tags[item_index])
            #     plt.show()
        self.assert_in_range(filename, train_fluxes)
        np.all(
            train_tags != train_tags.astype(int)
        ), "train_tags contains integer values!"
        return [train_fluxes, cadences], train_tags

    def assert_in_range(self, object_id, array, values_range=(0, 1)):
        if np.isnan(array).any():
            raise ValueError("Target " + str(object_id) + " contains NaN values")
        elif np.max(array) >= values_range[1]:
            raise ValueError("Target " + str(object_id) + " contains values > 1")
        elif np.min(array) < values_range[0]:
            raise ValueError("Target " + str(object_id) + " contains values < 0")
        # elif np.all(array == values_range[0]):
        #     raise ValueError("Target " + str(object_id) + " contains all values == 0")

    def on_epoch_end(self):
        if self.shuffle:
            self.generator_info_df = shuffle(self.generator_info_df)

    def class_weights(self):
        return {0: 1, 1: 1}

    def steps_per_epoch(self):
        return self.steps_count



import numpy

G = 6.67384e-11  # gravitational constant [m^3 / kg / s^2]
R_sun = 695508000  # radius of the Sun [m]
R_earth = 6371000  # radius of the Earth [m]
R_jup = 69911000  # radius of Jupiter [m]
M_sun = 1.989 * 10 ** 30  # mass of the Sun [kg]
SECONDS_PER_DAY = 86400


def t14(R_s, M_s, P, R_p=None, small_planet=False):
    """Planetary transit duration assuming a central transit on a circular orbit

    Parameters
    ----------
    R_s : float
        Stellar radius (in solar radii)
    M_s : float
        Stellar mass (in solar masses)
    P : float
        Planetary period (in days)
    small_planet : bool, default: `False`
        Small planet assumption if `True` (zero planetary radius). Otherwise, a
        planetary radius of 2 Jupiter radii is assumed (maximum reasonable planet size)

    Returns
    -------
    t14 : float
        Planetary transit duration (in days)
    """
    if R_p is not None:
        planet_size = R_sun * R_p
    elif small_planet:
        planet_size = 0
    else:
        planet_size = 2 * R_jup
    t14 = (
                  (R_sun * R_s + planet_size)
                  * (
                          (4 * P * SECONDS_PER_DAY)
                          / (numpy.pi * G * M_sun * M_s)
                  )
                  ** (1 / 3)
          ) / SECONDS_PER_DAY
    return t14


import matplotlib.pyplot as plt
import foldedleastsquares
from numpy.random import default_rng
from exoml.ete6.ete6_generator import Ete6ModelGenerator
from exoml.ml.encoding.time_position import value_encode_times
from sklearn.utils import shuffle
import numpy as np
import pandas as pd


class IatsonModelGenerator(Ete6ModelGenerator):
    TYPE_TO_LABEL = {'bckEB': [1, 0, 0, 0], 'EB': [0, 1, 0, 0], 'planet': [0, 0, 1, 0], 'none': [0, 0, 0, 1]}

    def __init__(self, injected_objects_df, lcs_dir, batch_size, input_sizes, zero_epsilon=1e-7,
                 measurements_per_point=13):
        super().__init__(zero_epsilon)
        self.injected_objects_df = injected_objects_df
        self.lcs_dir = lcs_dir
        self.batch_size = batch_size
        self.input_sizes = input_sizes
        self.random_number_generator = default_rng()
        self.measurements_per_point = measurements_per_point

    def __len__(self):
        return (np.ceil(len(self.injected_objects_df) / float(self.batch_size))).astype(int)

    def class_weights(self):
        return {0: 1, 1: 1, 2: 1, 3: 1}

    def _plot_df(self, df, type, scenario):
        fig, axs = plt.subplots(2, 3, figsize=(12, 6), constrained_layout=True)
        axs[0][0].scatter(df['#time'], df['flux'])
        axs[0][1].scatter(df['#time'], df['centroid_x'])
        axs[0][2].scatter(df['#time'], df['centroid_y'])
        axs[1][0].scatter(df['#time'], df['bck_flux'])
        axs[1][1].scatter(df['#time'], df['motion_x'])
        axs[1][2].scatter(df['#time'], df['motion_y'])
        plt.title(type + " " + scenario)
        plt.show()
        plt.clf()
        plt.close()

    def _plot_input(self, input_array, type, scenario):
        fig, axs = plt.subplots(4, 3, figsize=(12, 6), constrained_layout=True)
        current_array = np.transpose(input_array)[1]
        axs[0][0].scatter(np.transpose(input_array)[0], current_array)
        axs[0][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[3]
        axs[0][1].scatter(np.transpose(input_array)[0], current_array)
        axs[0][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[4]
        axs[0][2].scatter(np.transpose(input_array)[0], current_array)
        axs[0][2].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[5]
        axs[1][0].scatter(np.transpose(input_array)[0], current_array)
        axs[1][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[6]
        axs[1][1].scatter(np.transpose(input_array)[0], current_array)
        axs[1][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[7]
        axs[1][2].scatter(np.transpose(input_array)[0], current_array)
        axs[1][2].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[8]
        axs[2][0].scatter(np.transpose(input_array)[0], current_array)
        axs[2][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[9]
        axs[2][1].scatter(np.transpose(input_array)[0], current_array)
        axs[2][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[10]
        axs[2][2].scatter(np.transpose(input_array)[0], current_array)
        axs[2][2].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[11]
        axs[3][0].scatter(np.transpose(input_array)[0], current_array)
        axs[3][0].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        current_array = np.transpose(input_array)[12]
        axs[3][1].scatter(np.transpose(input_array)[0], current_array)
        axs[3][1].set_ylim(np.mean(current_array) - 6 * np.std(current_array), np.mean(current_array) + 6 * np.std(current_array))
        plt.title(type + " " + scenario)
        plt.show()
        plt.clf()
        plt.close()

    def __getitem__(self, idx):
        max_index = (idx + 1) * self.batch_size
        max_index = len(self.injected_objects_df) if max_index > len(self.injected_objects_df) else max_index
        target_indexes = np.arange(idx * self.batch_size, max_index, 1)
        target_indexes = shuffle(target_indexes)
        injected_objects_df = self.injected_objects_df.iloc[target_indexes]
        star_array = np.empty((len(target_indexes), 11, 1))
        star_neighbours_array = np.empty((len(target_indexes), 9 * 15, 1))
        global_flux_array = np.empty((len(target_indexes), 2500, self.measurements_per_point))
        folded_flux_even_array = np.empty((len(target_indexes), 500, self.measurements_per_point))
        folded_flux_odd_array = np.empty((len(target_indexes), 500, self.measurements_per_point))
        folded_flux_even_subhar_array = np.empty((len(target_indexes), 500, self.measurements_per_point))
        folded_flux_odd_subhar_array = np.empty((len(target_indexes), 500, self.measurements_per_point))
        folded_flux_even_har_array = np.empty((len(target_indexes), 500, self.measurements_per_point))
        folded_flux_odd_har_array = np.empty((len(target_indexes), 500, self.measurements_per_point))
        batch_data_values = np.empty((len(target_indexes), len(self.TYPE_TO_LABEL)))
        i = 0
        for df_index, target_row in injected_objects_df.iterrows():
            target_id = int(target_row['TIC ID'])
            #target_id = 281734251
            # target_row = self.injected_objects_df[self.injected_objects_df['TIC ID'] == 356110099].iloc[0]
            leading_zeros_object_id = '{:09}'.format(target_id)
            lc_filename = str(leading_zeros_object_id) + '_lc.csv'
            star_filename = str(leading_zeros_object_id) + '_star.csv'
            star_neighbours_df = pd.read_csv(self.lcs_dir + '/' + star_filename,
                                             usecols=['Teff', 'lum', 'v', 'j', 'k', 'h', 'radius', 'mass',
                                                      'dist_arcsec'], index_col=False)
            star_neighbours_df = self._prepare_input_neighbour_stars(star_neighbours_df)
            star_df = pd.read_csv(self.lcs_dir + '/' + star_filename, index_col=False)
            star_df = self._prepare_input_star(star_df)
            lc_df = pd.read_csv(self.lcs_dir + '/' + lc_filename, usecols=['#time', 'flux', 'flux_0', 'flux_1',
                                                                           'flux_2', 'flux_3', 'flux_4',
                                                                           'flux_err', 'centroid_x',
                                                                           'centroid_y', 'motion_x', 'motion_y',
                                                                           'bck_flux'], low_memory=True)
            lc_df = self._prepare_input_lc(lc_df)
            type = target_row['type']
            batch_data_values[i] = self.TYPE_TO_LABEL[type]
            not_null_times_args = np.argwhere(lc_df['#time'].to_numpy() > 0).flatten()
            lc_df = lc_df.iloc[not_null_times_args]
            period = target_row['period']
            epoch = target_row['epoch']
            duration = target_row['duration(h)'] / 24
            duration_to_period = duration / period
            #['ld_a', 'ld_b', 'Teff', 'lum', 'logg', 'radius', 'mass', 'v', 'j', 'h', 'k'])
            neighbours_array = star_neighbours_df.to_numpy().flatten()
            star_neighbours_array[i] = np.transpose([neighbours_array if len(neighbours_array) == 9 * 15 \
                else neighbours_array + np.zeros(9 * 15 - len(neighbours_array))])
            star_array[i] = np.transpose([star_df.to_numpy()])
            time = lc_df['#time'].to_numpy()
            # Global flux
            # Shifting data 1/4 so that main transit and possible occultation don't get cut by the borders
            lc_df_sorted_fold = lc_df.copy()
            lc_df_sorted_fold['#time'] = foldedleastsquares.fold(time, period, epoch + period / 4)
            lc_df_sorted_fold = lc_df_sorted_fold.sort_values(by=['#time'])
            lc_df_sorted_fold = lc_df_sorted_fold.sort_values(by=['#time'])
            global_flux_array[i] = self.bin_by_time(lc_df_sorted_fold.to_numpy(), self.input_sizes[2])
            #self._plot_input(global_flux_array[i], type, "even")
            # Focus flux even
            lc_df_focus = lc_df.copy()
            lc_df_focus['#time'] = foldedleastsquares.fold(time, period, epoch)
            lc_df_focus = lc_df_focus.sort_values(by=['#time'])
            lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 2) &
                                      (lc_df_focus['#time'] < 0.5 + duration_to_period * 2)]
            folded_flux_even_array[i] = self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[3])
            # self._plot_df(lc_df_focus, type, "even")
            # self._plot_input(folded_flux_even_array[i], type, "even")
            # Focus flux odd
            lc_df_focus = lc_df.copy()
            lc_df_focus['#time'] = foldedleastsquares.fold(time, period, epoch + period / 2)
            lc_df_focus = lc_df_focus.sort_values(by=['#time'])
            lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                      (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
            folded_flux_odd_array[i] = self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[4])
            # self._plot_df(lc_df_focus, type, "odd")
            # self._plot_input(folded_flux_odd_array[i], type, "odd")
            # Focus flux sub-harmonic even
            lc_df_focus = lc_df.copy()
            lc_df_focus['#time'] = foldedleastsquares.fold(time, period / 2, epoch)
            lc_df_focus = lc_df_focus.sort_values(by=['#time'])
            lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                      (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
            folded_flux_even_subhar_array[i] = self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[5])
            # Focus flux sub-harmonic odd
            lc_df_focus = lc_df.copy()
            lc_df_focus['#time'] = foldedleastsquares.fold(time, period / 2, epoch + period / 4)
            lc_df_focus = lc_df_focus.sort_values(by=['#time'])
            lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                      (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
            folded_flux_odd_subhar_array[i] = self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[6])
            # Focus flux harmonic even
            lc_df_focus = lc_df.copy()
            lc_df_focus['#time'] = foldedleastsquares.fold(time, period * 2, epoch)
            lc_df_focus = lc_df_focus.sort_values(by=['#time'])
            lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                      (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
            folded_flux_even_har_array[i] = self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[7])
            # Focus flux harmonic odd
            lc_df_focus = lc_df.copy()
            lc_df_focus['#time'] = foldedleastsquares.fold(time, period * 2, epoch + period * 3 / 2)
            lc_df_focus = lc_df_focus.sort_values(by=['#time'])
            lc_df_focus = lc_df_focus[(lc_df_focus['#time'] > 0.5 - duration_to_period * 3) &
                                      (lc_df_focus['#time'] < 0.5 + duration_to_period * 3)]
            folded_flux_odd_har_array[i] = self.bin_by_time(lc_df_focus.to_numpy(), self.input_sizes[8])
            assert np.max(global_flux_array[i]) < 2
            assert not np.isnan(star_array[i]).any() and not np.isinf(star_array[i]).any()
            assert not np.isnan(global_flux_array[i]).any() and not np.isinf(global_flux_array[i]).any()
            assert not np.isnan(folded_flux_even_array[i]).any() and not np.isinf(folded_flux_even_array[i]).any()
            assert not np.isnan(folded_flux_odd_array[i]).any() and not np.isinf(folded_flux_odd_array[i]).any()
            assert not np.isnan(folded_flux_even_subhar_array[i]).any() and not np.isinf(folded_flux_even_subhar_array[i]).any()
            assert not np.isnan(folded_flux_odd_subhar_array[i]).any() and not np.isinf(folded_flux_odd_subhar_array[i]).any()
            assert not np.isnan(folded_flux_even_har_array[i]).any() and not np.isinf(folded_flux_even_har_array[i]).any()
            assert not np.isnan(folded_flux_odd_har_array[i]).any() and not np.isinf(folded_flux_odd_har_array[i]).any()
            # logging.info("GENERATOR Star Inputs max " + str(np.max(star_array[i])) + " and min " +
            #              str(np.min(star_array[i])))
            # logging.info("GENERATOR Global Flux Inputs max " + str(np.max(global_flux_array[i])) + " and min " +
            #              str(np.min(global_flux_array[i])))
            # logging.info("GENERATOR Folded Even Inputs max " + str(np.max(folded_flux_even_array[i])) + " and min " +
            #              str(np.min(folded_flux_even_array[i])))
            # logging.info("GENERATOR Folded Odd Inputs max " + str(np.max(folded_flux_odd_array[i])) + " and min " +
            #              str(np.min(folded_flux_odd_array[i])))
            # logging.info("GENERATOR Folded Even Subhar Inputs max " + str(np.max(folded_flux_even_subhar_array[i])) +
            #              " and min " + str(np.min(folded_flux_even_subhar_array[i])))
            # logging.info("GENERATOR Folded Odd Subhar Inputs max " + str(np.max(folded_flux_odd_subhar_array[i])) +
            #              " and min " + str(np.min(folded_flux_odd_subhar_array[i])))
            # logging.info("GENERATOR Folded Even Har Inputs max " + str(np.max(folded_flux_even_har_array[i])) +
            #              " and min " + str(np.min(folded_flux_even_har_array[i])))
            # logging.info("GENERATOR Folded Odd Har Inputs max " + str(np.max(folded_flux_odd_har_array[i])) +
            #              " and min " + str(np.min(folded_flux_odd_har_array[i])))
            i = i + 1
        return [star_array, star_neighbours_array, global_flux_array, folded_flux_even_array, folded_flux_odd_array,
                folded_flux_even_subhar_array, folded_flux_odd_subhar_array, folded_flux_even_har_array,
                folded_flux_odd_har_array], batch_data_values
#
# df = pd.read_csv("~/Downloads/injected_objects.csv", index_col=False)
# df = shuffle(df)
# df.reset_index(inplace=True, drop=True)
# img = IatsonModelGenerator(df, "~/Downloads/", 1, [11, 9 * 15, 2500, 500, 500, 500, 500, 500, 500])
# img.__getitem__(1)



import logging

from exoml.ete6.ete6_generator import Ete6ModelGenerator
from exoml.ml.encoding.time_position import value_encode_times
from sklearn.utils import shuffle
import tensorflow as tf
import numpy as np
import pandas as pd


class DetrendModelGenerator(Ete6ModelGenerator):
    """
    Sequence generator for DETREND model batches
    """
    def __init__(self, lc_filenames, batch_size, input_size, zero_epsilon):
        super().__init__(zero_epsilon)
        self.lc_filenames = lc_filenames
        self.batch_size = batch_size
        self.input_size = input_size

    def __len__(self):
        return (np.ceil(len(self.lc_filenames) / float(self.batch_size))).astype(int)

    def __getitem__(self, idx):
        batch_filenames = self.lc_filenames[idx * self.batch_size: (idx + 1) * self.batch_size]
        filenames_shuffled = shuffle(batch_filenames)
        batch_data_array = np.empty((len(filenames_shuffled), self.input_size[0], self.input_size[1]))
        batch_data_values = np.empty((len(filenames_shuffled), self.input_size[0]))
        i = 0
        for file in filenames_shuffled:
            input_df = pd.read_csv(file, usecols=['#time', 'flux', 'flux_err', 'centroid_x', 'centroid_y',
                                                  'motion_x', 'motion_y', 'bck_flux'], low_memory=True)
            values_df = pd.read_csv(file, usecols=['eb_model', 'bckeb_model', 'planet_model'],
                                    low_memory=True)
            input_df = self.__prepare_input_data(input_df)
            batch_data_array[i] = input_df.to_numpy()
            assert not np.isnan(batch_data_array[i]).any() and not np.isinf(batch_data_array[i]).any()
            values_df['model'] = 1 - ((1 - values_df['eb_model']) + (1 - values_df['bckeb_model']) +
                                      (1 - values_df['planet_model']))
            batch_data_values[i] = values_df['model'].to_numpy()
            batch_data_values[i] = np.nan_to_num(batch_data_values[i], nan=self.zero_epsilon)
            negative_values_args = np.argwhere(batch_data_values[i] <= 0).flatten()
            batch_data_values[i][negative_values_args] = self.zero_epsilon
            assert not np.isnan(batch_data_values[i]).any() and not np.isinf(batch_data_values[i]).any()
            logging.info("GENERATOR Inputs max " + str(np.max(batch_data_array[i])) + " and min " +
                         str(np.min(batch_data_array[i])))
            logging.info("GENERATOR Values max " + str(np.max(batch_data_values[i])) + " and min " +
                         str(np.min(batch_data_values[i])))
            i = i + 1
        return batch_data_array, batch_data_values

    def __prepare_input_data(self, input_df):
        time = input_df["#time"].to_numpy()
        input_df["#time"] = value_encode_times(time - time[0])
        dif = time[1:] - time[:-1]
        jumps = np.where(np.abs(dif) > 0.25)[0]
        jumps = np.append(jumps, len(input_df))
        previous_jump_index = 0
        for jumpIndex in jumps:
            token = input_df["centroid_x"][previous_jump_index:jumpIndex].to_numpy()
            input_df["centroid_x"][previous_jump_index:jumpIndex] = (np.tanh(token - np.nanmedian(token)) + 1) / 2
            token = input_df["centroid_y"][previous_jump_index:jumpIndex].to_numpy()
            input_df["centroid_y"][previous_jump_index:jumpIndex] = (np.tanh(token - np.nanmedian(token)) + 1) / 2
            token = input_df["motion_x"][previous_jump_index:jumpIndex].to_numpy()
            input_df["motion_x"][previous_jump_index:jumpIndex] = (np.tanh(token - np.nanmedian(token)) + 1) / 2
            token = input_df["motion_y"][previous_jump_index:jumpIndex].to_numpy()
            input_df["motion_y"][previous_jump_index:jumpIndex] = (np.tanh(token - np.nanmedian(token)) + 1) / 2
            token = input_df["bck_flux"][previous_jump_index:jumpIndex].to_numpy()
            input_df["bck_flux"][previous_jump_index:jumpIndex] = (np.tanh(token - np.nanmedian(token)) + 1) / 3
            previous_jump_index = jumpIndex
        input_df = input_df.fillna(self.zero_epsilon)
        input_df = input_df.replace(0.0, self.zero_epsilon)
        input_df = input_df.replace(0, self.zero_epsilon)
        return input_df

    def class_weights(self):
        return None



import logging
import threading

import numpy as np
import pandas as pd
from sklearn.utils import shuffle
from tensorflow.keras.utils import Sequence
from matplotlib import pyplot as plt
from functools import lru_cache


@lru_cache(maxsize=5)  # keep up to 5 files cached
def load_flux(path, delimiter=','):
    return np.loadtxt(path, delimiter=delimiter)


class SantoGenerator(Sequence):
    def __init__(
            self,
            lcs_dir,
            target_files,
            input_size=500,
            step_size=1,
            batch_size=500,
            shuffle=True,
            zero_epsilon=1e-7,
            indexes_steps=1,
            plot=False,
            from_arrays=False,
            store_arrays=False
    ):
        self.lcs_dir = lcs_dir
        self.target_files = target_files
        self.input_size = input_size
        self.step_size = step_size
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.zero_epsilon = zero_epsilon
        self.plot = plot
        self.indexes_steps = indexes_steps
        self.from_arrays = from_arrays
        self.store_arrays = store_arrays
        self.steps_count, self.generator_info_df = self._count_inputs()

    def _count_inputs(self):
        return self.__len__(), len(self.target_files)

    def __len__(self):
        return len(self.target_files) // self.batch_size

    def __getitem__(self, index):
        train_fluxes = np.full((self.batch_size, self.input_size), 0.0)
        train_tags = np.full((self.batch_size, self.input_size), float(0))
        cadences = np.full((self.batch_size), 0.0)
        initial_index = index * self.batch_size
        for iteration_index in np.arange(initial_index, initial_index + self.batch_size):
            item_index = iteration_index - initial_index
            filename = self.target_files[iteration_index]
            try:
                flux = load_flux(f"{filename}", delimiter=",")
                flux_data = flux[np.r_[1, 0]]
                tags_data = flux[3]
                flux_data[0] = flux_data[0] / 2
                train_fluxes[item_index] = flux_data[0]
                train_tags[item_index] = tags_data
                cadences[item_index] = np.median(flux_data[1][np.argwhere(flux_data[1] > 0).flatten()]) * 24
                cadences[item_index] = np.where(cadences[item_index] >= 1, 1 - self.zero_epsilon, cadences[item_index])
                self.assert_in_range("", flux_data[0])
            except Exception as e:
                logging.exception(
                    f"Error with file {filename}. file_index {item_index}, "
                    f"iteration_index {iteration_index}, batch_size {self.batch_size}, "
                    f"input_size {self.input_size}"
                )
                raise e
            # if self.plot:
            #     fig, ax = plt.subplots(2, 1, figsize=(16, 8), constrained_layout=True)
            #     ax[0].scatter(np.arange(0, len(train_fluxes[item_index])), train_fluxes[item_index])
            #     ax[1].plot(np.arange(0, train_tags[item_index]), train_tags[item_index])
            #     plt.show()
        np.all(
            train_tags != train_tags.astype(int)
        ), "train_tags contains integer values!"
        return [train_fluxes, cadences], train_tags

    def assert_in_range(self, object_id, array, values_range=(0, 1)):
        if np.isnan(array).any():
            raise ValueError("Target " + str(object_id) + " contains NaN values")
        elif np.max(array) >= values_range[1]:
            raise ValueError("Target " + str(object_id) + " contains values > 1")
        elif np.min(array) < values_range[0]:
            raise ValueError("Target " + str(object_id) + " contains values < 0")
        # elif np.all(array == values_range[0]):
        #     raise ValueError("Target " + str(object_id) + " contains all values == 0")

    def on_epoch_end(self):
        if self.shuffle:
            self.target_files = shuffle(self.target_files)

    def class_weights(self):
        return {0: 1, 1: 1}

    def steps_per_epoch(self):
        return self.steps_count



import multiprocessing

import numpy as np
import pandas as pd
import keras
from keras.layers import LeakyReLU
from keras.utils import losses_utils
from sklearn.utils import shuffle

from exoml.iatson.iatson_generator import IatsonModelGenerator
from exoml.ml.model.base_model import CategoricalPredictionSetStats
from exoml.ml.model.imbalanced_categorical_model import ImbalancedCategoricalModel


class IATSON(ImbalancedCategoricalModel):
    def __init__(self, class_ids, class_names, dropout_rate=0.1) -> None:
        super().__init__(class_names, 'IATSON', [11, 9 * 15, 2500, 500, 500, 500, 500, 500, 500], class_ids, dropout_rate)
        self.class_names = class_names

    def _get_flux_conv_branch(self, name):
        # (time, flux, [detrended_flux1 ... detrended_flux5], flux_model, centroidx, centroidy, motionx, motiony, bck)
        # flux model by transit params and stellar params
        leaky_relu_alpha = 0.01
        flux_input = keras.Input(shape=(2500, 13), name=name)
        flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_input)
        flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=512, kernel_size=5, padding="same",
                                          activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=5, strides=3)(flux_branch)
        flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
        flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=256, kernel_size=5, padding="same",
                                          activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=5, strides=2)(flux_branch)
        flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
        flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=128, kernel_size=5, padding="same",
                                          activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=5, strides=2)(flux_branch)
        flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
        flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=64, kernel_size=4, padding="same",
                                          activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=4, strides=2)(flux_branch)
        flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
        flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=32, kernel_size=3, padding="same",
                                          activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=3, strides=2)(flux_branch)
        flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(flux_branch)
        flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(flux_branch)
        flux_branch = keras.layers.Conv1D(filters=16, kernel_size=3, padding="same",
                                          activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = keras.layers.MaxPooling1D(pool_size=3, strides=2)(flux_branch)
        return flux_input, flux_branch

    def _get_flux_model_branch(self):
        leaky_relu_alpha = 0.01
        flux_input, flux_branch = self._get_flux_conv_branch("global_flux_branch")
        flux_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha))(flux_branch)
        flux_branch = keras.layers.Dropout(rate=self.dropout_rate)(flux_branch)
        flux_branch = keras.layers.Dense(16, activation='linear')(flux_branch)
        flux_branch = keras.layers.LayerNormalization()(flux_branch)
        flux_branch = keras.Model(inputs=[flux_input], outputs=flux_branch)
        return flux_branch

    def _get_focus_flux_branch(self, name):
        leaky_relu_alpha = 0.01
        focus_flux_input = keras.Input(shape=(500, 13), name=name)
        focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_input)
        # (time, flux, detrended_flux1... detrended_flux5, flux_model, centroidx, centroidy, motionx, motiony, bck)
        # flux model by transit params and stellar params
        focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=256, kernel_size=10, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=10, strides=2)(focus_flux_branch)
        focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=128, kernel_size=6, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=6, strides=2)(focus_flux_branch)
        focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=64, kernel_size=3, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=3, strides=2)(focus_flux_branch)
        focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=32, kernel_size=3, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=3, strides=2)(focus_flux_branch)
        focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=16, kernel_size=2, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=2, strides=2)(focus_flux_branch)
        focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.SpatialDropout1D(rate=self.dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Conv1D(filters=9, kernel_size=2, padding="same",
                                                activation=LeakyReLU(leaky_relu_alpha), use_bias=True)(
            focus_flux_branch)
        focus_flux_branch = keras.layers.MaxPooling1D(pool_size=2, strides=2)(focus_flux_branch)
        focus_flux_branch = keras.layers.normalization.batch_normalization.BatchNormalization()(focus_flux_branch)
        focus_flux_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha))(focus_flux_branch)
        focus_flux_branch = keras.layers.Dropout(rate=self.dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.LayerNormalization()(focus_flux_branch)
        return focus_flux_input, focus_flux_branch

    def _get_focus_flux_model_branch(self):
        leaky_relu_alpha = 0.01
        odd_flux_input, odd_flux_branch = self._get_focus_flux_branch("focus_odd_flux_branch")
        even_flux_input, even_flux_branch = self._get_focus_flux_branch("focus_even_flux_branch")
        harmonic_odd_flux_input, harmonic_odd_flux_branch = self._get_focus_flux_branch("focus_harmonic_odd_flux_branch")
        harmonic_even_flux_input, harmonic_even_flux_branch = self._get_focus_flux_branch("focus_harmonic_even_flux_branch")
        subharmonic_odd_flux_input, subharmonic_odd_flux_branch = self._get_focus_flux_branch(
            "focus_subharmonic_odd_flux_branch")
        subharmonic_even_flux_input, subharmonic_even_flux_branch = self._get_focus_flux_branch(
            "focus_subharmonic_even_flux_branch")
        odd_flux_branch = keras.layers.Add()(
            [odd_flux_branch, harmonic_odd_flux_branch, subharmonic_odd_flux_branch])
        even_flux_branch = keras.layers.Add()(
            [even_flux_branch, harmonic_even_flux_branch, subharmonic_even_flux_branch])
        odd_flux_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha))(odd_flux_branch)
        odd_flux_branch = keras.layers.Dropout(rate=self.dropout_rate)(odd_flux_branch)
        even_flux_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha))(even_flux_branch)
        even_flux_branch = keras.layers.Dropout(rate=self.dropout_rate)(even_flux_branch)
        odd_flux_branch = keras.layers.Dense(32, activation='linear')(odd_flux_branch)
        even_flux_branch = keras.layers.Dense(32, activation='linear')(even_flux_branch)
        odd_flux_branch = keras.layers.LayerNormalization()(odd_flux_branch)
        even_flux_branch = keras.layers.LayerNormalization()(even_flux_branch)
        focus_flux_branch = keras.layers.Add()([odd_flux_branch, even_flux_branch])
        focus_flux_branch = keras.layers.Dense(24, activation=LeakyReLU(leaky_relu_alpha))(focus_flux_branch)
        focus_flux_branch = keras.layers.Dropout(rate=self.dropout_rate)(focus_flux_branch)
        focus_flux_branch = keras.layers.Dense(16, activation='linear')(focus_flux_branch)
        input = [even_flux_input, odd_flux_input, subharmonic_even_flux_input, subharmonic_odd_flux_input,
                 harmonic_even_flux_input, harmonic_odd_flux_input]
        focus_flux_branch = keras.Model(inputs=input, outputs=focus_flux_branch)
        return focus_flux_branch

    def _get_singletransit_tpf_model(self):
        video_image_width = 13
        video_image_height = 13
        video_image_channels = 1
        sequences_per_video = 100
        tpf_model_input = keras.Input(
            shape=(video_image_height, video_image_width, sequences_per_video, video_image_channels),
            name="tpf_input")
        tpf_model = keras.layers.SpatialDropout3D(rate=0.3)(tpf_model_input)
        tpf_model = keras.layers.Conv3D(filters=100, kernel_size=(3, 3, 5), strides=(1, 1, 3), activation="relu")(
            tpf_model)
        tpf_model = keras.layers.SpatialDropout3D(rate=0.2)(tpf_model)
        tpf_model = keras.layers.Conv3D(filters=200, kernel_size=(3, 3, 5), strides=(1, 1, 3), activation="relu")(
            tpf_model)
        tpf_model = keras.layers.SpatialDropout3D(rate=0.1)(tpf_model)
        tpf_model = keras.layers.MaxPooling3D(pool_size=(5, 5, 10), strides=(3, 3, 6), padding='same')(tpf_model)
        tpf_model = keras.layers.Dense(200, activation="relu")(tpf_model)
        tpf_model = keras.layers.Dense(100, activation="relu")(tpf_model)
        tpf_model = keras.layers.Dense(20, activation="relu")(tpf_model)
        tpf_model = keras.layers.Flatten()(tpf_model)
        return keras.Model(inputs=tpf_model_input, outputs=tpf_model)

    def _get_singletransit_motion_centroids_model(self):
        mc_input = keras.Input(
            shape=(100, 4),
            name="motion_centroids_input")
        mc_model = keras.layers.Conv1D(filters=50, kernel_size=3, strides=3, activation="relu", use_bias=True,
                                       padding='same')(mc_input)
        mc_model = keras.layers.SpatialDropout1D(rate=0.3)(mc_model)
        mc_model = keras.layers.Conv1D(filters=100, kernel_size=5, strides=5, activation="relu", use_bias=True,
                                       padding='same')(mc_model)
        mc_model = keras.layers.SpatialDropout1D(rate=0.2)(mc_model)
        mc_model = keras.layers.MaxPooling1D(pool_size=5, strides=3, padding='same')(mc_model)
        mc_model = keras.layers.Dense(50, activation="relu")(mc_model)
        mc_model = keras.layers.Dense(20, activation="relu")(mc_model)
        mc_model = keras.layers.Flatten()(mc_model)
        return keras.Model(inputs=mc_input, outputs=mc_model)

    def _get_singletransit_bckflux_model(self):
        bck_input = keras.Input(shape=(100, 1), name="bck_input")
        bck_model = keras.layers.Conv1D(filters=25, kernel_size=2, strides=2, activation="relu", use_bias=True,
                                        padding='same')(bck_input)
        bck_model = keras.layers.SpatialDropout1D(rate=0.3)(bck_model)
        bck_model = keras.layers.Conv1D(filters=50, kernel_size=3, strides=3, activation="relu", use_bias=True,
                                        padding='same')(bck_model)
        bck_model = keras.layers.SpatialDropout1D(rate=0.2)(bck_model)
        bck_model = keras.layers.MaxPooling1D(pool_size=5, strides=3, padding='same')(bck_model)
        bck_model = keras.layers.Dense(50, activation="relu")(bck_model)
        bck_model = keras.layers.Dense(10, activation="relu")(bck_model)
        bck_model = keras.layers.Flatten()(bck_model)
        return keras.Model(inputs=bck_input, outputs=bck_model)

    def _get_singletransit_flux_model(self):
        bck_input = keras.Input(shape=(100, 1), name="flux_input")
        bck_model = keras.layers.Conv1D(filters=25, kernel_size=2, strides=2, activation="relu", use_bias=True,
                                        padding='same')(bck_input)
        bck_model = keras.layers.SpatialDropout1D(rate=0.3)(bck_model)
        bck_model = keras.layers.Conv1D(filters=50, kernel_size=3, strides=3, activation="relu", use_bias=True,
                                        padding='same')(bck_model)
        bck_model = keras.layers.SpatialDropout1D(rate=0.2)(bck_model)
        bck_model = keras.layers.MaxPooling1D(pool_size=5, strides=3, padding='same')(bck_model)
        bck_model = keras.layers.Dense(50, activation="relu")(bck_model)
        bck_model = keras.layers.Dense(10, activation="relu")(bck_model)
        bck_model = keras.layers.Flatten()(bck_model)
        return keras.Model(inputs=bck_input, outputs=bck_model)

    def _get_single_transit_model(self):
        tpf_branch = self._get_singletransit_tpf_model()
        mc_branch = self._get_singletransit_motion_centroids_model()
        bck_branch = self._get_singletransit_bckflux_model()
        flux_branch = self._get_singletransit_flux_model()
        final_branch = keras.layers.concatenate(
            [tpf_branch.output, mc_branch.output, bck_branch.output, flux_branch.output], axis=1)
        final_branch = keras.layers.Dense(64, activation="relu", name="final-dense1")(final_branch)
        final_branch = keras.layers.Dense(32, activation="relu", name="final-dense2")(final_branch)
        final_branch = keras.layers.Dense(1, activation="softmax", name="final-dense-softmax")(final_branch)
        inputs = tpf_branch.inputs + mc_branch.inputs + bck_branch.inputs + flux_branch.inputs
        model = keras.Model(inputs=inputs, outputs=final_branch, name="mnist_model")
        keras.utils.vis_utils.plot_model(model, "IATSON_model.png", show_shapes=True)

    def build(self):
        leaky_relu_alpha = 0.01
        stellar_model_input = keras.Input(shape=(11, 1), name="stellar_model")
        stellar_model_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha), name="stellar-first")(
            stellar_model_input)
        stellar_model_branch = keras.layers.Dropout(rate=self.dropout_rate, name="stellar-first-dropout-0.1")(stellar_model_branch)
        stellar_model_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha), name="stellar-refinement")(
            stellar_model_branch)
        #(TEFF, lum, 4 magnitudes, distance, radius, mass) * 15 stars
        stellar_neighbours_input = keras.Input(shape=(9 * 15, 1), name="stellar_neighbours_model")
        stellar_neighbours_branch = keras.layers.Dense(100, activation=LeakyReLU(leaky_relu_alpha), name="stellar-neighbours-1")(
            stellar_neighbours_input)
        stellar_neighbours_branch = keras.layers.Dropout(rate=self.dropout_rate, name="stellar-neighbours-1-dropout-0.1")(stellar_neighbours_branch)
        stellar_neighbours_branch = keras.layers.Dense(50, activation=LeakyReLU(leaky_relu_alpha), name="stellar-neighbours-2")(
            stellar_neighbours_branch)
        stellar_neighbours_branch = keras.layers.Dropout(rate=self.dropout_rate, name="stellar-neighbours-2-dropout-0.1")(stellar_neighbours_branch)
        stellar_neighbours_branch = keras.layers.Dense(16, activation=LeakyReLU(leaky_relu_alpha), name="stellar-neighbours_3")(
            stellar_neighbours_branch)
        flux_model_branch = self._get_flux_model_branch()
        focus_flux_model_branch = self._get_focus_flux_model_branch()
        final_branch = keras.layers.Concatenate(axis=1)(
            [stellar_model_branch, stellar_neighbours_branch, flux_model_branch.output, focus_flux_model_branch.output])
        final_branch = keras.layers.GlobalAveragePooling1D()(final_branch)
        final_branch = keras.layers.Dense(64, activation=LeakyReLU(leaky_relu_alpha), name="final-dense1")(final_branch)
        final_branch = keras.layers.Dropout(rate=self.dropout_rate)(final_branch)
        final_branch = keras.layers.Dense(32, activation=LeakyReLU(leaky_relu_alpha), name="final-dense2")(final_branch)
        final_branch = keras.layers.Dropout(rate=self.dropout_rate)(final_branch)
        final_branch = keras.layers.Dense(4, activation="softmax", name="final-dense-softmax")(final_branch)
        inputs = [stellar_model_input, stellar_neighbours_input] + flux_model_branch.inputs + focus_flux_model_branch.inputs
        self.set_model(keras.Model(inputs=inputs, outputs=final_branch, name="IATSON"))
        return self

    def load_training_set(self, **kwargs):
        training_dir = kwargs.get('training_dir')
        injected_objects_df = pd.read_csv(training_dir + "/injected_objects.csv", index_col=None)
        injected_objects_df = shuffle(injected_objects_df)
        injected_objects_df.reset_index(inplace=True, drop=True)
        return injected_objects_df

    def instance_generator(self, dataset, dir, batch_size, input_sizes, zero_epsilon):
        return IatsonModelGenerator(dataset, dir, batch_size, input_sizes, zero_epsilon)

    def predict(self, input):
        prediction = self.model(input, training=False)
        return self.predict_threshold([prediction])

    def predict_batch(self, inputs, expected_outputs=None, batch_size=20, cores=multiprocessing.cpu_count() - 1):
        predictions = self.model.predict(inputs, use_multiprocessing=True, batch_size=batch_size, workers=cores)
        max_prediction_indexes, max_prediction_values = self.predict_threshold(predictions)
        if expected_outputs is not None:
            bckeb_stats = self.test_metrics(expected_outputs, predictions, 0)
            eb_stats = self.test_metrics(expected_outputs, predictions, 1)
            planet_stats = self.test_metrics(expected_outputs, predictions, 2)
            none_stats = self.test_metrics(expected_outputs, predictions, 3)
            return max_prediction_indexes, max_prediction_values, [bckeb_stats, eb_stats, planet_stats, none_stats]
        else:
            return max_prediction_indexes, max_prediction_values

    def predict_threshold(self, predictions, threshold=[0.5, 0.5, 0.5, 0.5]):
        max_prediction_indexes = [np.argmax(prediction) for prediction in predictions]
        max_prediction_values = [predictions[index][max_prediction_index]
                                 if predictions[index][max_prediction_index] > threshold[predictions[index][max_prediction_index]]
                                 else -1
                                 for index, max_prediction_index in enumerate(max_prediction_indexes)]
        max_prediction_indexes[max_prediction_values == -1] = -1
        return max_prediction_indexes, max_prediction_values

    def predict_test_set(self, training_dir, model_dir, batch_size=20, test_set_limit=None, zero_epsilon=1e-7):
        test_dataset = pd.read_csv(model_dir + '/test_dataset.csv')
        test_dataset = shuffle(test_dataset)
        if test_set_limit is not None:
            test_dataset = test_dataset.iloc[0:test_set_limit]
        lcs_dir = training_dir + '/lcs/'
        testing_batch_generator = IatsonModelGenerator(test_dataset, lcs_dir, batch_size, self.input_size,
                                                       zero_epsilon)
        expected_outputs = None
        for i in np.arange(0, len(test_dataset) // batch_size + 1, 1):
            input, expected_output = testing_batch_generator.__getitem__(i)
            expected_outputs = np.concatenate((expected_outputs, expected_output)) if expected_outputs is not None else expected_output
        return self.predict_batch(testing_batch_generator, expected_outputs, batch_size)

    def test_metrics(self, expected_outputs, max_prediction_indexes, label_index):
        set_count = len(expected_outputs)
        max_output_indexes = [np.max(expected_output) for expected_output in expected_outputs]
        matches_indexes = np.argwhere(max_output_indexes == max_prediction_indexes).flatten()
        matches_count = len(matches_indexes)
        accuracy = matches_count / set_count
        expected_label_indexes = np.argwhere(max_output_indexes == label_index).flatten()
        output_label_indexes = np.argwhere(max_output_indexes == label_index).flatten()
        label_tp_indexes = np.intersect1d(expected_label_indexes, output_label_indexes)
        label_fp_indexes = output_label_indexes[~np.isin(output_label_indexes, expected_label_indexes)]
        label_fn_indexes = expected_label_indexes[~np.isin(expected_label_indexes, output_label_indexes)]
        label_tp = len(label_tp_indexes)
        label_fp = len(label_fp_indexes)
        label_fn = len(label_fn_indexes)
        precision = label_tp / (label_tp + label_fp)
        recall = label_tp / (label_tp + label_fn)
        return CategoricalPredictionSetStats(label_tp, label_fp, label_fn, accuracy, precision, recall)



import warnings

from keras.callbacks import ModelCheckpoint
import tensorflow as tf
import numpy as np
from keras.src import backend


def sync_to_numpy_or_python_type(tensors):
    """Took this method from abandonned tensorflow_addons

    Syncs and converts a structure of `Tensor`s to `NumPy` arrays or Python
    scalar types.

    For each tensor, it calls `tensor.numpy()`. If the result is a scalar value,
    it converts it to a Python type, such as a float or int, by calling
    `result.item()`.

    Numpy scalars are converted, as Python types are often more convenient to
    deal with. This is especially useful for bfloat16 Numpy scalars, which don't
    support as many operations as other Numpy values.

    Async strategies (such as `TPUStrategy` and `ParameterServerStrategy`) are
    forced to
    sync during this process.

    Args:
      tensors: A structure of tensors.

    Returns:
      `tensors`, but scalar tensors are converted to Python types and non-scalar
      tensors are converted to Numpy arrays.
    """
    if isinstance(tensors, tf.distribute.experimental.coordinator.RemoteValue):
        tensors = tensors.fetch()

    def _to_single_numpy_or_python_type(t):
        # Don't turn ragged or sparse tensors to NumPy.
        if isinstance(t, tf.Tensor):
            t = t.numpy()
        # Strings, ragged and sparse tensors don't have .item(). Return them
        # as-is.
        if not isinstance(t, (np.ndarray, np.generic)):
            return t
        return t.item() if np.ndim(t) == 0 else t

    return tf.nest.map_structure(_to_single_numpy_or_python_type, tensors)


class ModelCheckpointCallback(ModelCheckpoint):
    def __init__(self, filepath, original_path, monitor: str = "val_loss", verbose: int = 0,
                 save_best_only: bool = False,
                 save_weights_only: bool = False, mode: str = "auto", save_freq="epoch",
                 initial_value_threshold=None, filter_metrics: list[str] = []):
        super().__init__(filepath=filepath,
                         monitor=monitor, verbose=verbose, save_best_only=save_best_only,
                         save_weights_only=save_weights_only, mode=mode, save_freq=save_freq,
                         initial_value_threshold=initial_value_threshold)
        self.original_path = original_path
        self.filter_metrics = filter_metrics
        self.epochs_since_last_save = 0

    def on_epoch_end(self, epoch, logs=None):
        self.epochs_since_last_save += 1
        if self.save_freq == "epoch":
            self._save_model(epoch=epoch, batch=None, logs=logs)

    def _should_save_model(self, epoch, batch, logs, filepath):
        """Determines whether the model should be saved.

        The model should be saved in the following cases:

        - self.save_best_only is False
        - self.save_best_only is True and `monitor` is a numpy array or
          backend tensor (falls back to `save_best_only=False`)
        - self.save_best_only is True and `self.monitor_op(current, self.best)`
          evaluates to True.

        Args:
            epoch: the epoch this iteration is in.
            batch: the batch this iteration is in. `None` if the `save_freq`
                is set to `"epoch"`.
            logs: the `logs` dict passed in to `on_batch_end` or
                `on_epoch_end`.
            filepath: the path where the model would be saved
        """
        logs = logs or {}
        if self.save_best_only:
            current = logs.get(self.monitor)
            if current is None:
                warnings.warn(
                    f"Can save best model only with {self.monitor} available.",
                    stacklevel=2,
                )
                return True
            elif (
                isinstance(current, np.ndarray) or backend.is_tensor(current)
            ) and len(current.shape) > 0:
                warnings.warn(
                    "Can save best model only when `monitor` is "
                    f"a scalar value. Received: {current}. "
                    "Falling back to `save_best_only=False`."
                )
                return True
            else:
                if self.monitor_op(current, self.best) and np.all(np.array(self.filter_metrics) > 0):
                    self.best = current
                    return True
                else:
                    return False
        else:
            return True





from keras.callbacks import Callback
import threading


class MetricsPlotCallback(Callback):

    def __init__(self, base_model, model_dir, steps_per_epoch):
        super().__init__()
        self.base_model = base_model
        self.model_dir = model_dir
        self.steps_per_epoch = steps_per_epoch

    def on_epoch_end(self, epoch, logs=None):
        # thread = threading.Thread(target=self.base_model.plot_metrics,
        #                           args=(self.model_dir, self.steps_per_epoch, True))
        # thread.start()
        self.base_model.plot_metrics(self.model_dir, self.steps_per_epoch, True)
